# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import argparse
import logging
import sys

import yaml

from django.core import exceptions
from django.core.management.base import BaseCommand

from promgen import models

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Import http probes from blackbox_exporter configuration"
    field = models.Probe._meta.get_field("description")

    # Reuse input validation from Django's createsuperuser
    def get_input_data(self, field, message, default=None):
        raw_value = input(message)
        if default and raw_value == "":
            raw_value = default
        try:
            val = field.clean(raw_value, None)
        except exceptions.ValidationError as e:
            self.stderr.write("Error: %s" % "; ".join(e.messages))
            val = None

        return val

    def add_arguments(self, parser):
        parser.add_argument(
            "config",
            help="Path to blackbox-exporter configuration",
            type=argparse.FileType("r"),
            default=sys.stdin,
        )

    def handle(self, config, verbosity, **kwargs):
        logging.root.setLevel(
            {
                0: logging.ERROR,
                1: logging.WARNING,
                2: logging.INFO,
                3: logging.DEBUG,
            }.get(verbosity)
        )
        try:
            config = yaml.safe_load(config)

            # See online example for file format
            # https://github.com/prometheus/blackbox_exporter/blob/master/example.yml

            for name, module in config.get("modules", {}).items():
                if module["prober"] != "http":
                    logger.debug("Skipping non HTTP: %s %s", name, module["prober"])
                    continue

                m, created = models.Probe.objects.get_or_create(module=name)

                if created:
                    while m.description == "":
                        m.description = self.get_input_data(
                            self.field, "Please enter description for %s: " % name
                        )
                    m.save(update_fields=["description"])
                    self.stdout.write("Added probe %s" % m)
        except KeyboardInterrupt:
            self.stderr.write("\nOperation cancelled.")
            sys.exit(1)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from promgen import models, util

from django.conf import settings


def settings_in_view(request):
    return {
        "TIMEZONE": util.setting("timezone", "UTC"),
        "VERSION": settings.PROMGEN_VERSION,
        "DEFAULT_EXPORTERS": models.DefaultExporter.objects.order_by("job", "-port"),
    }



# Generated by Django 2.2.2 on 2019-08-09 02:06

from django.db import migrations


def to_promgen(apps, schema_editor):
    Rule = apps.get_model("promgen", "Rule")
    ContentType = apps.get_model("contenttypes", "ContentType")
    sites, _ = ContentType.objects.get_or_create(app_label="sites", model="site")
    promgen, _ = ContentType.objects.get_or_create(app_label="promgen", model="site")
    Rule.objects.filter(content_type_id=sites.id).update(content_type_id=promgen.id)


def to_sites(apps, schema_editor):
    Rule = apps.get_model("promgen", "Rule")
    ContentType = apps.get_model("contenttypes", "ContentType")
    sites, _ = ContentType.objects.get_or_create(app_label="sites", model="site")
    promgen, _ = ContentType.objects.get_or_create(app_label="promgen", model="site")
    Rule.objects.filter(content_type_id=promgen.id).update(content_type_id=sites.id)


class Migration(migrations.Migration):
    dependencies = [
        ("promgen", "0009_blackbox-module"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [migrations.RunPython(to_promgen, to_sites)]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from unittest import mock

from django.core.exceptions import ValidationError
from django.test import override_settings
from django.urls import reverse

import promgen.templatetags.promgen as macro
from promgen import models, prometheus, tests, views

_RULE_V2 = """
groups:
- name: promgen.example.com
  rules:
  - alert: example-rule
    annotations:
      rule: https://promgen.example.com/rule/1
      summary: Example rule summary
    expr: up==1
    for: 1s
    labels:
      severity: high
""".lstrip().encode(
    "utf-8"
)

TEST_SETTINGS = tests.Data("examples", "promgen.yml").yaml()


class RuleTest(tests.PromgenTest):
    fixtures = ["testcases.yaml", "extras.yaml"]

    @override_settings(PROMGEN_SCHEME="https")
    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_write_new(self, mock_post):
        result = prometheus.render_rules()
        self.assertEqual(result, _RULE_V2)

    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_copy(self, mock_post):
        rule = models.Rule.objects.get(pk=1)
        copy = rule.copy_to(content_type="service", object_id=2)
        # Test that our copy has the same labels and annotations
        self.assertIn("severity", copy.labels)
        self.assertIn("summary", copy.annotations)

    @override_settings(PROMGEN=TEST_SETTINGS)
    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_import_v2(self, mock_post):
        self.user = self.force_login(username="demo")
        self.add_user_permissions("promgen.change_rule", "promgen.change_site")
        response = self.client.post(
            reverse("rule-import"),
            {"rules": tests.Data("examples", "import.rule.yml").raw()},
            follow=True,
        )

        # Includes count of our setUp rule + imported rules
        self.assertRoute(response, views.RuleImport, status=200)
        self.assertCount(models.Rule, 3, "Missing Rule")

    @override_settings(PROMGEN=TEST_SETTINGS)
    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_import_project_rule(self, mock_post):
        self.user = self.force_login(username="demo")
        self.add_user_permissions("promgen.add_rule", "promgen.change_project")

        response = self.client.post(
            reverse("rule-new", kwargs={"content_type": "project", "object_id": 1}),
            {"rules": tests.Data("examples", "import.rule.yml").raw()},
            follow=True,
        )
        self.assertRoute(response, views.ProjectDetail, status=200)
        self.assertCount(models.Rule, 3, "Missing Rule")

    @override_settings(PROMGEN=TEST_SETTINGS)
    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_import_service_rule(self, mock_post):
        self.user = self.force_login(username="demo")
        self.add_user_permissions("promgen.add_rule", "promgen.change_service")
        response = self.client.post(
            reverse(
                "rule-new",
                kwargs={"content_type": "service", "object_id": 1},
            ),
            {"rules": tests.Data("examples", "import.rule.yml").raw()},
            follow=True,
        )
        self.assertRoute(response, views.ServiceDetail, status=200)
        self.assertCount(models.Rule, 3, "Missing Rule")

    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_missing_permission(self, mock_post):
        self.client.post(
            reverse("rule-import"),
            {"rules": tests.Data("examples", "import.rule.yml").raw()},
        )

        # Should only be a single rule from our initial setup
        self.assertCount(models.Rule, 1, "Missing Rule")

    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_macro(self, mock_post):
        self.site = models.Site.objects.get(pk=1)
        self.service = models.Service.objects.get(pk=1)
        self.project = models.Project.objects.get(pk=1)

        clause = "up{%s}" % macro.EXCLUSION_MACRO

        rules = {
            "common": {"assert": 'up{service!~"test-service"}'},
            "service": {"assert": 'up{service="test-service",project!~"test-project"}'},
            "project": {"assert": 'up{service="test-service",project="test-project",}'},
        }

        common_rule = models.Rule.objects.create(
            name="Common", clause=clause, duration="1s", obj=self.site
        )
        rules["common"]["model"] = models.Rule.objects.get(pk=common_rule.pk)
        service_rule = common_rule.copy_to("service", self.service.id)
        rules["service"]["model"] = models.Rule.objects.get(pk=service_rule.pk)
        project_rule = service_rule.copy_to("project", self.project.id)
        rules["project"]["model"] = models.Rule.objects.get(pk=project_rule.pk)

        for k, r in rules.items():
            self.assertEqual(macro.rulemacro(r["model"]), r["assert"], "Expansion wrong for %s" % k)

    @override_settings(PROMGEN=TEST_SETTINGS)
    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_invalid_annotation_value(self, mock_post):
        rule = models.Rule.objects.get(pk=1)
        # $label.foo is invalid (should be $labels) so make sure we raise an exception
        rule.annotations["summary"] = "{{$label.foo}}"
        with self.assertRaises(ValidationError):
            prometheus.check_rules([rule])

    @override_settings(PROMGEN=TEST_SETTINGS)
    @mock.patch("django.dispatch.dispatcher.Signal.send")
    def test_invalid_annotation_name(self, mock_post):
        rule = models.Rule.objects.get(pk=1)
        # $label.foo is invalid (should be $labels) so make sure we raise an exception
        rule.annotations["has a space"] = "value"
        with self.assertRaises(ValidationError):
            prometheus.check_rules([rule])



# Generated by Django 2.1.2 on 2018-11-16 02:23

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('promgen', '0004_merge_20180905_0247'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='owner',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='service',
            name='owner',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 2.2.4 on 2019-11-28 02:21

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0010_app_label_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='alert',
            name='error_count',
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name='alert',
            name='sent_count',
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.CreateModel(
            name='AlertError',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.TextField()),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('alert', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Alert')),
            ],
        ),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE
import collections
import logging
import os
from urllib.parse import urljoin

from atomicwrites import atomic_write
from celery import shared_task
from requests.exceptions import RequestException

from promgen import models, prometheus, util, notification

logger = logging.getLogger(__name__)


@shared_task
def index_alert(alert_pk):
    alert = models.Alert.objects.get(pk=alert_pk)
    labels = alert.json.get("commonLabels")
    for name, value in labels.items():
        models.AlertLabel.objects.create(alert=alert, name=name, value=value)


@shared_task
def process_alert(alert_pk):
    """
    Process alert for routing and notifications

    We load our Alert from the database and expand it to determine which labels are routable

    Next we loop through all senders configured and de-duplicate sender:target pairs before
    queueing the notification to actually be sent
    """
    alert = models.Alert.objects.get(pk=alert_pk)
    routable, data = alert.expand()

    # For any blacklisted label patterns, we delete them from the queue
    # and consider it done (do not send any notification)
    blacklist = util.setting("alertmanager:blacklist", {})
    for key in blacklist:
        logger.debug("Checking key %s", key)
        if key in data["commonLabels"]:
            if data["commonLabels"][key] in blacklist[key]:
                logger.debug("Blacklisted label %s", blacklist[key])
                alert.delete()
                return

    # After processing our blacklist, it should be safe to queue our
    # alert to also index the labels
    index_alert.delay(alert.pk)

    # Now that we have our routable items, we want to check which senders are
    # configured and expand those as needed
    senders = collections.defaultdict(set)
    for label, obj in routable.items():
        logger.debug("Processing %s %s", label, obj)
        for sender in models.Sender.objects.filter(obj=obj, enabled=True):
            if sender.filtered(data):
                logger.debug("Filtered out sender %s", sender)
                continue
            if hasattr(sender.driver, "splay"):
                for splay in sender.driver.splay(sender.value):
                    senders[splay.sender].add(splay.value)
            else:
                senders[sender.sender].add(sender.value)

    for driver in senders:
        for target in senders[driver]:
            send_notification.delay(driver, target, data, alert_pk=alert.pk)


@shared_task
def send_alert(sender, target, data):
    """
    Send alert to specific target

    alert_pk is used when querying our alert normally and is missing
    when we send a test message. In the case we send a test message
    we want to raise any exceptions so that the test function can
    handle it
    """
    logger.debug("Sending %s %s", sender, target)

    try:
        notifier = notification.load(sender)
        notifier._send(target, data)
    except ImportError:
        logging.exception("Error loading plugin %s", sender)
        raise
    except RequestException:
        logging.exception("Error sending notification %s", sender)
        raise
    except Exception:
        logging.exception("Unknown Error")
        raise


@shared_task
def send_notification(*args, alert_pk, **kwargs):
    """
    Send notification to target

    This wraps send_alert, but wraps it so that we can keep track
    of the number of sent and error counts
    """
    try:
        send_alert(*args, **kwargs)
    except Exception as e:
        util.inc_for_pk(models.Alert, pk=alert_pk, error_count=1)
        models.AlertError.objects.create(alert_id=alert_pk, message=str(e))
    else:
        util.inc_for_pk(models.Alert, pk=alert_pk, sent_count=1)


@shared_task
def reload_prometheus():
    from promgen import signals

    target = urljoin(util.setting("prometheus:url"), "/-/reload")
    response = util.post(target)
    response.raise_for_status()
    signals.post_reload.send(response)


@shared_task
def clear_tombstones():
    target = urljoin(util.setting("prometheus:url"), "/api/v1/admin/tsdb/clean_tombstones")
    response = util.post(target)
    response.raise_for_status()


@shared_task
def write_urls(path=None, reload=True, chmod=0o644):
    if path is None:
        path = util.setting("prometheus:blackbox")
    with atomic_write(path, overwrite=True) as fp:
        # Set mode on our temporary file before we write and move it
        os.chmod(fp.name, chmod)
        fp.write(prometheus.render_urls())
    if reload:
        reload_prometheus()


@shared_task
def write_config(path=None, reload=True, chmod=0o644):
    if path is None:
        path = util.setting("prometheus:targets")
    with atomic_write(path, overwrite=True) as fp:
        # Set mode on our temporary file before we write and move it
        os.chmod(fp.name, chmod)
        fp.write(prometheus.render_config())
    if reload:
        reload_prometheus()


@shared_task
def write_rules(path=None, reload=True, chmod=0o644):
    if path is None:
        path = util.setting("prometheus:rules")
    with atomic_write(path, mode="wb", overwrite=True) as fp:
        # Set mode on our temporary file before we write and move it
        os.chmod(fp.name, chmod)
        fp.write(prometheus.render_rules())
    if reload:
        reload_prometheus()



# Generated by Django 2.1.2 on 2019-03-06 06:20

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0006_exporter_options'),
    ]

    operations = [
        migrations.CreateModel(
            name='Filter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('value', models.CharField(max_length=128)),
            ],
            options={
                'ordering': ('sender', 'name', 'value'),
            },
        ),
        migrations.AddField(
            model_name='filter',
            name='sender',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Sender'),
        ),
        migrations.AlterUniqueTogether(
            name='filter',
            unique_together={('sender', 'name', 'value')},
        ),
    ]



# Generated by Django 3.2.4 on 2021-07-06 06:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0020_fix_exporter_constraints'),
    ]

    operations = [
        migrations.AddField(
            model_name='shard',
            name='samples',
            field=models.PositiveBigIntegerField(default=5000000, help_text='Estimated Sample Count'),
        ),
        migrations.AddField(
            model_name='shard',
            name='targets',
            field=models.PositiveBigIntegerField(default=10000, help_text='Estimated Target Count'),
        ),
    ]



# Generated by Django 2.2.10 on 2020-03-16 09:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0018_default_exporter_scheme'),
    ]

    operations = [
        migrations.AddField(
            model_name='sender',
            name='enabled',
            field=models.BooleanField(default=True),
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-26 08:15

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0002_auto_20180316_0525'),
    ]

    operations = [
        migrations.CreateModel(
            name='Alert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(default=django.utils.timezone.now)),
                ('body', models.TextField()),
            ],
        ),
    ]



# Generated by Django 2.2.8 on 2020-02-28 06:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0016_alertlabel'),
    ]

    operations = [
        migrations.AlterField(
            model_name='alertlabel',
            name='value',
            field=models.TextField(),
        ),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import json

from django.core.management.base import BaseCommand

from promgen import prometheus, util
from promgen.signals import (
    trigger_write_config,
    trigger_write_rules,
    trigger_write_urls,
)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("target_file")
        parser.add_argument("replace_shard", nargs="?")

    def handle(self, target_file, replace_shard, **kwargs):
        if target_file.startswith("http"):
            config = util.get(target_file).json()
        else:
            config = json.load(open(target_file), encoding="utf8")

        imported, skipped = prometheus.import_config(config, replace_shard)

        if imported:
            counters = {key: len(imported[key]) for key in imported}
            self.stdout.write(f"Imported {counters}")

        if skipped:
            counters = {key: len(skipped[key]) for key in skipped}
            self.stdout.write(f"Skipped {counters}")

        trigger_write_config.send(self, force=True)
        trigger_write_rules.send(self, force=True)
        trigger_write_urls.send(self, force=True)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import json
import logging

import django.contrib.sites.models
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.forms.models import model_to_dict
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.text import slugify

import promgen.templatetags.promgen as macro
from promgen import plugins, tests, validators
from promgen.shortcuts import resolve_domain

logger = logging.getLogger(__name__)


class Site(django.contrib.sites.models.Site):
    # Proxy model for sites so that we can easily
    # query our related Rules
    rule_set = GenericRelation("promgen.Rule", for_concrete_model=False)

    def get_absolute_url(self):
        return reverse("site-detail")

    class Meta:
        proxy = True


class ObjectFilterManager(models.Manager):
    def create(self, *args, **kwargs):
        if "obj" in kwargs:
            obj = kwargs.pop("obj")
            kwargs["object_id"] = obj.id
            kwargs["content_type_id"] = ContentType.objects.get_for_model(obj).id
        return self.get_queryset().create(*args, **kwargs)

    def filter(self, *args, **kwargs):
        if "obj" in kwargs:
            obj = kwargs.pop("obj")
            kwargs["object_id"] = obj.id
            kwargs["content_type_id"] = ContentType.objects.get_for_model(obj).id
        return self.get_queryset().filter(*args, **kwargs)

    def get_or_create(self, *args, **kwargs):
        if "obj" in kwargs:
            obj = kwargs.pop("obj")
            kwargs["object_id"] = obj.id
            kwargs["content_type_id"] = ContentType.objects.get_for_model(obj).id
        if "defaults" in kwargs and "obj" in kwargs["defaults"]:
            obj = kwargs["defaults"].pop("obj")
            kwargs["defaults"]["object_id"] = obj.id
            kwargs["defaults"]["content_type_id"] = ContentType.objects.get_for_model(obj).id

        return self.get_queryset().get_or_create(*args, **kwargs)


class Sender(models.Model):
    objects = ObjectFilterManager()

    sender = models.CharField(max_length=128)
    value = models.CharField(max_length=128)
    alias = models.CharField(max_length=128, blank=True)

    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        limit_choices_to=(
            models.Q(app_label="auth", model="user")
            | models.Q(app_label="promgen", model="project")
            | models.Q(app_label="promgen", model="service")
        ),
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    owner = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True)

    enabled = models.BooleanField(default=True)

    def show_value(self):
        if self.alias:
            return self.alias
        return self.value

    show_value.short_description = "Value"

    def __str__(self):
        return f"{self.sender}:{self.show_value()}"

    @classmethod
    def driver_set(cls):
        """Return the list of drivers for Sender model"""
        for entry in plugins.notifications():
            try:
                yield entry.module_name, entry.load()
            except ImportError:
                logger.warning("Error importing %s", entry.module_name)

    __driver = {}

    @property
    def driver(self):
        """Return configured driver for Sender model instance"""
        if self.sender in self.__driver:
            return self.__driver[self.sender]

        for entry in plugins.notifications():
            try:
                self.__driver[entry.module_name] = entry.load()()
            except ImportError:
                logger.warning("Error importing %s", entry.module_name)
        return self.__driver[self.sender]

    def test(self):
        """
        Test sender plugin

        Uses the same test json from our unittests but subs in the currently
        tested object as part of the test data
        """
        data = tests.Data("examples", "alertmanager.json").json()
        if hasattr(self.content_object, "name"):
            data["commonLabels"][self.content_type.name] = self.content_object.name
            for alert in data.get("alerts", []):
                alert["labels"][self.content_type.name] = self.content_object.name

        from promgen import tasks

        tasks.send_alert(self.sender, self.value, data)

    def filtered(self, alert):
        """
        Check filters for a specific sender

        If no filters are defined, then we let the message through
        If filters are defined, then we check to see if at least one filter matches
        If no filters match, then we assume it's filtered out
        """
        logger.debug("Checking labels %s", alert["commonLabels"])
        # If we have no specific whitelist, then we let everything through
        if self.filter_set.count() == 0:
            return False

        # If we have filters defined, then we need to check to see if our
        # filters match
        for f in self.filter_set.all():
            logger.debug("Checking filter %s %s", f.name, f.value)
            if alert["commonLabels"].get(f.name) == f.value:
                return False
        # If none of our filters match, then we blacklist this sender
        return True


class Filter(models.Model):
    sender = models.ForeignKey("Sender", on_delete=models.CASCADE)
    name = models.CharField(max_length=128)
    value = models.CharField(max_length=128)

    class Meta:
        ordering = ("sender", "name", "value")
        unique_together = (("sender", "name", "value"),)


class Shard(models.Model):
    name = models.CharField(max_length=128, unique=True, validators=[validators.labelvalue])
    url = models.URLField(max_length=256)
    proxy = models.BooleanField(
        default=False,
        help_text="Queries can be proxied to these shards",
    )
    enabled = models.BooleanField(
        default=True,
        help_text="Able to register new Services and Projects",
    )

    samples = models.PositiveBigIntegerField(
        default=5000000,
        help_text="Estimated Sample Count",
    )
    targets = models.PositiveBigIntegerField(
        default=10000,
        help_text="Estimated Target Count",
    )

    class Meta:
        ordering = ["name"]

    def get_absolute_url(self):
        return reverse("datasource-detail", kwargs={"pk": self.pk})

    def __str__(self):
        if self.enabled:
            return self.name
        return self.name + " (disabled)"


class Service(models.Model):
    name = models.CharField(max_length=128, unique=True, validators=[validators.labelvalue])
    description = models.TextField(blank=True)
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, default=None
    )

    notifiers = GenericRelation(Sender)
    rule_set = GenericRelation("Rule")

    class Meta:
        ordering = ["name"]

    def get_absolute_url(self):
        return reverse("service-detail", kwargs={"pk": self.pk})

    def __str__(self):
        return self.name

    @classmethod
    def default(cls, service_name="Default", shard_name="Default"):
        shard, created = Shard.objects.get_or_create(name=shard_name)
        if created:
            logger.info("Created default shard")

        service, created = cls.objects.get_or_create(name=service_name, defaults={"shard": shard})
        if created:
            logger.info("Created default service")
        return service


class Project(models.Model):
    name = models.CharField(max_length=128, unique=True, validators=[validators.labelvalue])
    description = models.TextField(blank=True)
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, default=None
    )

    service = models.ForeignKey("promgen.Service", on_delete=models.CASCADE)
    shard = models.ForeignKey("promgen.Shard", on_delete=models.CASCADE)
    farm = models.ForeignKey("promgen.Farm", blank=True, null=True, on_delete=models.SET_NULL)

    notifiers = GenericRelation(Sender)
    rule_set = GenericRelation("Rule")

    class Meta:
        ordering = ["name"]

    def get_absolute_url(self):
        return reverse("project-detail", kwargs={"pk": self.pk})

    def __str__(self):
        return f"{self.service} » {self.name}"


class Farm(models.Model):
    name = models.CharField(max_length=128, validators=[validators.labelvalue])
    source = models.CharField(max_length=128)

    class Meta:
        ordering = ["name"]
        unique_together = (("name", "source"),)

    def get_absolute_url(self):
        return reverse("farm-detail", kwargs={"pk": self.pk})

    def refresh(self):
        target = set()
        current = {host.name for host in self.host_set.all()}
        for entry in plugins.discovery():
            if self.source == entry.name:
                target.update(entry.load()().fetch(self.name))

        remove = current - target
        add = target - current

        if add:
            Audit.log(f"Adding {add} to {self}", self)
            Host.objects.bulk_create([Host(name=name, farm_id=self.id) for name in add])

        if remove:
            Audit.log(f"Removing {add} from {self}", self)
            Host.objects.filter(farm=self, name__in=remove).delete()

        return add, remove

    @classmethod
    def fetch(cls, source):
        for entry in plugins.discovery():
            if entry.name == source:
                yield from entry.load()().farms()

    @cached_property
    def driver(self):
        """Return configured driver for Farm model instance"""
        for entry in plugins.discovery():
            if entry.name == self.source:
                return entry.load()()

    @property
    def editable(self):
        return not self.driver.remote

    @classmethod
    def driver_set(cls):
        """Return the list of drivers for Farm model"""
        for entry in plugins.discovery():
            yield entry.name, entry.load()()

    def __str__(self):
        return f"{self.name} ({self.source})"


class Host(models.Model):
    name = models.CharField(max_length=128)
    farm = models.ForeignKey("Farm", on_delete=models.CASCADE)

    class Meta:
        ordering = ["name"]
        unique_together = (("name", "farm"),)

    def get_absolute_url(self):
        return reverse("host-detail", kwargs={"slug": self.name})

    def __str__(self):
        return f"{self.name} [{self.farm.name}]"


class BaseExporter(models.Model):
    job = models.CharField(max_length=128, help_text="Exporter name. Example node, jmx, app")
    port = models.IntegerField(help_text="Port Exporter is running on")
    path = models.CharField(
        max_length=128, blank=True, help_text="Exporter path. Defaults to /metrics"
    )
    scheme = models.CharField(
        max_length=5,
        choices=(("http", "http"), ("https", "https")),
        default="http",
        help_text="Scrape exporter over http or https",
    )

    class Meta:
        abstract = True


class DefaultExporter(BaseExporter):
    class Meta:
        ordering = ["job", "port"]
        unique_together = (("job", "port", "path"),)


class Exporter(BaseExporter):
    project = models.ForeignKey("Project", on_delete=models.CASCADE)
    enabled = models.BooleanField(default=True)

    class Meta:
        ordering = ["job", "port"]
        unique_together = (("job", "port", "path", "scheme", "project"),)

    def __str__(self):
        return "{}:{}{}".format(self.job, self.port, self.path or "/metrics")


class Probe(models.Model):
    module = models.CharField(
        help_text="Probe Module from blackbox_exporter config", max_length=128, unique=True
    )
    description = models.TextField(blank=True)

    def __str__(self):
        return f"{self.module} » {self.description}"


class URL(models.Model):
    url = models.URLField(max_length=256)
    project = models.ForeignKey("Project", on_delete=models.CASCADE)
    probe = models.ForeignKey("promgen.Probe", on_delete=models.CASCADE)

    class Meta:
        ordering = ["project__service", "project", "url"]

    def __str__(self):
        return f"{self.project} [{self.url}]"


class Rule(models.Model):
    objects = ObjectFilterManager()

    name = models.CharField(max_length=128, unique=True, validators=[validators.metricname])
    clause = models.TextField(help_text="Prometheus query")
    duration = models.CharField(
        max_length=128,
        validators=[validators.duration],
        help_text="Duration field with postfix. Example 30s, 5m, 1d",
    )
    enabled = models.BooleanField(default=True)
    parent = models.ForeignKey(
        "Rule", null=True, related_name="overrides", on_delete=models.SET_NULL
    )

    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        limit_choices_to=(
            models.Q(app_label="promgen", model="site")
            | models.Q(app_label="promgen", model="project")
            | models.Q(app_label="promgen", model="service")
        ),
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id", for_concrete_model=False)
    description = models.TextField(blank=True)

    labels = models.JSONField(default=dict)
    annotations = models.JSONField(default=dict)

    class Meta:
        ordering = ["content_type", "object_id", "name"]

    def __str__(self):
        return (
            f"{self.name}"
            if self.content_object is None
            else f"{self.name} [{self.content_object.name}]"
        )

    def get_absolute_url(self):
        return reverse("rule-detail", kwargs={"pk": self.pk})

    def set_object(self, content_type, object_id):
        self.content_type = ContentType.objects.get(
            model=content_type,
            app_label="promgen",
        )
        self.object_id = object_id

    def copy_to(self, content_type, object_id):
        """
        Make a copy under a new service

        It's important that we set pk to None so a new object is created, but we
        also need to ensure the new name is unique by appending some unique data
        to the end of the name
        """
        with transaction.atomic():
            content_type = ContentType.objects.get(model=content_type, app_label="promgen")

            # First check to see if this rule is already overwritten
            for rule in Rule.objects.filter(
                parent_id=self.pk, content_type=content_type, object_id=object_id
            ):
                return rule

            content_object = content_type.get_object_for_this_type(pk=object_id)

            orig_pk = self.pk
            self.pk = None
            self.parent_id = orig_pk
            self.name = f"{self.name}_{slugify(content_object.name)}".replace("-", "_")
            self.content_type = content_type
            self.object_id = object_id
            # Enable the copy by default since it's more likely the user prefers
            # to have their own copy enabled rather than the original one.
            self.enabled = True
            self.clause = self.clause.replace(
                macro.EXCLUSION_MACRO,
                f'{content_type.model}="{content_object.name}",{macro.EXCLUSION_MACRO}',
            )

            # Add a label to our new rule by default, to help ensure notifications
            # get routed to the notifier we expect
            self.labels[content_type.model] = content_object.name

            self.save()

        return self


class AlertLabel(models.Model):
    alert = models.ForeignKey("Alert", on_delete=models.CASCADE)
    name = models.CharField(max_length=128)
    value = models.TextField()


class Alert(models.Model):
    created = models.DateTimeField(default=timezone.now)
    body = models.TextField()
    sent_count = models.PositiveIntegerField(default=0)
    error_count = models.PositiveIntegerField(default=0)

    def get_absolute_url(self):
        return reverse("alert-detail", kwargs={"pk": self.pk})

    def expand(self):
        # Map of Prometheus labels to Promgen objects
        LABEL_MAPPING = [
            ("project", Project),
            ("service", Service),
        ]
        routable = {}
        data = json.loads(self.body)

        data.setdefault("commonLabels", {})
        data.setdefault("commonAnnotations", {})
        # Set our link back to Promgen for processed notifications
        # The original externalURL can still be visible from the alerts page
        data["externalURL"] = resolve_domain(self.get_absolute_url())

        # Look through our labels and find the object from Promgen's DB
        # If we find an object in Promgen, add an annotation with a direct link
        for label, klass in LABEL_MAPPING:
            if label not in data["commonLabels"]:
                logger.debug("Missing label %s", label)
                continue

            # Should only find a single value, but I think filter is a little
            # bit more forgiving than get in terms of throwing errors
            for obj in klass.objects.filter(name=data["commonLabels"][label]):
                logger.debug("Found %s %s", label, obj)
                routable[label] = obj
                data["commonAnnotations"][label] = resolve_domain(obj)

        return routable, data

    @cached_property
    def json(self):
        return json.loads(self.body)


class AlertError(models.Model):
    alert = models.ForeignKey(Alert, on_delete=models.CASCADE)
    created = models.DateTimeField(default=timezone.now)
    message = models.TextField()


class Audit(models.Model):
    body = models.TextField()
    created = models.DateTimeField()
    data = models.TextField(blank=True)
    old = models.TextField(blank=True)

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    object_id = models.PositiveIntegerField(default=0)
    content_object = GenericForeignKey("content_type", "object_id")

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True, default=None
    )

    @property
    def highlight(self):
        if self.body.startswith("Created"):
            return "success"
        if self.body.startswith("Updated"):
            return "warning"
        if self.body.startswith("Deleted"):
            return "danger"
        return ""

    @classmethod
    def log(cls, body, instance=None, old=None, **kwargs):
        from promgen.middleware import get_current_user

        kwargs["body"] = body
        kwargs["created"] = timezone.now()
        kwargs["user"] = get_current_user()

        if instance:
            kwargs["content_type"] = ContentType.objects.get_for_model(instance)
            kwargs["object_id"] = instance.id
            data_dict = model_to_dict(instance)
            if isinstance(instance, Sender) and data_dict["alias"]:
                data_dict["value"] = "********"
            kwargs["data"] = json.dumps(data_dict, sort_keys=True)
        if old:
            kwargs["old"] = json.dumps(model_to_dict(old), sort_keys=True)

        return cls.objects.create(**kwargs)


class Prometheus(models.Model):
    shard = models.ForeignKey("promgen.Shard", on_delete=models.CASCADE)
    host = models.CharField(max_length=128)
    port = models.IntegerField()

    def __str__(self):
        return f"{self.host}:{self.port}"

    class Meta:
        ordering = ["shard", "host"]
        unique_together = (("host", "port"),)
        verbose_name_plural = "prometheis"



# Generated by Django 2.0.7 on 2018-09-05 02:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0003_alert_queue'),
        ('promgen', '0003_default-group'),
    ]

    operations = [
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import os
import shutil

from django.conf import settings
from django.core.management.base import BaseCommand

from promgen import PROMGEN_CONFIG_DIR, PROMGEN_CONFIG_FILE

PROMGEN_CONFIG_DEFAULT = settings.BASE_DIR / "promgen" / "tests" / "examples" / "promgen.yml"


class Command(BaseCommand):
    # We manually run the system checks at the end
    requires_system_checks = False

    def prompt(self, prompt, *args, **kwargs):
        return input(prompt.format(*args, **kwargs))

    def write(self, str, color):
        self.stdout.write(color(str))

    def success(self, fmtstr, *args, **kwargs):
        self.write(fmtstr.format(*args, **kwargs), color=self.style.SUCCESS)

    def warning(self, fmtstr, *args, **kwargs):
        self.write(fmtstr.format(*args, **kwargs), color=self.style.WARNING)

    def setting(self, key, default=None, value=None):
        path = PROMGEN_CONFIG_DIR / key
        if path.exists():
            self.success("{:20} : {}", key, os.environ[key])
            return

        if default:
            if self.prompt("Use {} for {} ? (yes/no) ", default, key).lower() == "yes":
                value = default

        while not value:
            value = self.prompt("Please enter a value for {}: ", key).strip()

        self.warning("{:20} : {}", key, value)
        with path.open("w", encoding="utf8") as fp:
            fp.write(value)

    def handle(self, **kwargs):
        self.write("Bootstrapping Promgen", color=self.style.MIGRATE_HEADING)

        if not PROMGEN_CONFIG_DIR.exists():
            self.warning("Config {} Created", PROMGEN_CONFIG_DIR)
            os.makedirs(PROMGEN_CONFIG_DIR)

        if not PROMGEN_CONFIG_FILE.exists():
            self.warning(
                "Creating promgen config {} from {}",
                PROMGEN_CONFIG_FILE,
                PROMGEN_CONFIG_DEFAULT,
            )
            shutil.copy(PROMGEN_CONFIG_DEFAULT, PROMGEN_CONFIG_FILE)
        else:
            self.success("Config {} Exists", PROMGEN_CONFIG_FILE)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import logging

from django.core.management.base import BaseCommand

from promgen import prometheus, tasks

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--reload",
            action="store_true",
            help="Trigger Prometheus Reload",
        )
        parser.add_argument(
            "out",
            nargs="?",
            help="Optionally specify an output file to use an atomic write operation",
        )

    def handle(self, **kwargs):
        if kwargs["out"]:
            tasks.write_rules(
                path=kwargs["out"],
                reload=kwargs["reload"],
            )
        else:
            # Since we're already working with utf8 encoded data, we can skip
            # the newline ending here
            self.stdout.ending = None
            self.stdout.buffer.write(prometheus.render_rules())



# Generated by Django 1.11.10 on 2018-02-07 06:56

from django.conf import settings
import django.contrib.sites.models
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import promgen.models
import promgen.validators


class Migration(migrations.Migration):

    replaces = [('promgen', '0001_initial'), ('promgen', '0002_audit'), ('promgen', '0003_setting'), ('promgen', '0004_auto_20161019_0755'), ('promgen', '0005_sender'), ('promgen', '0006_auto_20161019_1214'), ('promgen', '0007_auto_20161020_0253'), ('promgen', '0008_auto_20161020_0528'), ('promgen', '0009_auto_20161021_0640'), ('promgen', '0010_auto_20161025_0034'), ('promgen', '0011_auto_20161025_0737'), ('promgen', '0012_auto_20161027_0306'), ('promgen', '0013_auto_20161027_0307'), ('promgen', '0014_stat'), ('promgen', '0015_delete_stat'), ('promgen', '0016_auto_20161121_0714'), ('promgen', '0017_auto_20161215_2352'), ('promgen', '0018_rule_enabled'), ('promgen', '0019_url'), ('promgen', '0020_auto_20161226_0337'), ('promgen', '0021_auto_20161228_0049'), ('promgen', '0022_auto_20170116_0824'), ('promgen', '0023_shard'), ('promgen', '0024_auto_20170220_0324'), ('promgen', '0025_shard_url'), ('promgen', '0026_auto_20170315_0344'), ('promgen', '0027_prometheus_server'), ('promgen', '0028_url_sort_fields'), ('promgen', '0029_auto_20170406_0236'), ('promgen', '0030_exporter_enabled'), ('promgen', '0031_sender_to_notification'), ('promgen', '0032_audit_diff'), ('promgen', '0033_diff_old'), ('promgen', '0034_auto_20170622_0518'), ('promgen', '0035_rule_parent'), ('promgen', '0036_auto_20170626_0231'), ('promgen', '0037_shard_proxy'), ('promgen', '0038_audit_user'), ('promgen', '0039_auto_20170725_0257'), ('promgen', '0040_default_farm'), ('promgen', '0041_sender_owner'), ('promgen', '0042_description_field'), ('promgen', '0043_rule_description'), ('promgen', '0044_common-rules')]

    initial = True

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Audit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('body', models.TextField()),
                ('created', models.DateTimeField()),
                ('data', models.TextField(blank=True)),
                ('old', models.TextField(blank=True)),
                ('object_id', models.PositiveIntegerField(default=0)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('user', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Exporter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('job', models.CharField(max_length=128)),
                ('port', models.IntegerField()),
                ('path', models.CharField(blank=True, max_length=128)),
                ('enabled', models.BooleanField(default=True)),
            ],
            options={
                'ordering': ['job', 'port'],
            },
        ),
        migrations.CreateModel(
            name='Farm',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('source', models.CharField(max_length=128)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Host',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('farm', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Farm')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True)),
                ('description', models.TextField(blank=True)),
                ('farm', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='promgen.Farm')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Prometheus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('host', models.CharField(max_length=128)),
                ('port', models.IntegerField()),
            ],
            options={
                'verbose_name_plural': 'prometheis',
                'ordering': ['shard', 'host'],
            },
        ),
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True, validators=[django.core.validators.RegexValidator('^[0-9a-zA-Z_]*$', 'Only alphanumeric characters are allowed.')])),
                ('clause', models.TextField(help_text='Prometheus query')),
                ('duration', models.CharField(help_text='Duration field with postfix. Example 30s, 5m, 1d', max_length=128, validators=[promgen.validators.duration])),
                ('enabled', models.BooleanField(default=True)),
                ('object_id', models.PositiveIntegerField()),
                ('description', models.TextField(blank=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('parent', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='overrides', to='promgen.Rule')),
            ],
            options={
                'ordering': ['content_type', 'object_id', 'name'],
            },
        ),
        migrations.CreateModel(
            name='RuleAnnotation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('value', models.TextField()),
                ('rule', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Rule')),
            ],
        ),
        migrations.CreateModel(
            name='RuleLabel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('value', models.CharField(max_length=128)),
                ('rule', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Rule')),
            ],
        ),
        migrations.CreateModel(
            name='Sender',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender', models.CharField(max_length=128)),
                ('value', models.CharField(max_length=128)),
                ('alias', models.CharField(blank=True, max_length=128)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('owner', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Service',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True)),
                ('description', models.TextField(blank=True)),
            ],
            options={
                'ordering': ['shard', 'name'],
            },
        ),
        migrations.CreateModel(
            name='Shard',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True)),
                ('url', models.URLField(max_length=256)),
                ('proxy', models.BooleanField(default=False)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='URL',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.URLField(max_length=256)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Project')),
            ],
            options={
                'ordering': ['project__service', 'project', 'url'],
            },
        ),
        migrations.CreateModel(
            name='Site',
            fields=[
            ],
            options={
                'indexes': [],
                'proxy': True,
            },
            bases=('sites.site',),
            managers=[
                ('objects', django.contrib.sites.models.SiteManager()),
            ],
        ),
        migrations.AddField(
            model_name='service',
            name='shard',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Shard'),
        ),
        migrations.AddField(
            model_name='prometheus',
            name='shard',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Shard'),
        ),
        migrations.AddField(
            model_name='project',
            name='service',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Service'),
        ),
        migrations.AlterUniqueTogether(
            name='farm',
            unique_together={('name', 'source')},
        ),
        migrations.AddField(
            model_name='exporter',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='promgen.Project'),
        ),
        migrations.AlterUniqueTogether(
            name='prometheus',
            unique_together={('host', 'port')},
        ),
        migrations.AlterUniqueTogether(
            name='host',
            unique_together={('name', 'farm')},
        ),
        migrations.AlterUniqueTogether(
            name='exporter',
            unique_together={('job', 'port', 'project')},
        ),
    ]



# Copyright (c) 2018 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import json
import logging

from django.core.management.base import BaseCommand
from django.test import override_settings

from promgen import models, tasks, tests


class Command(BaseCommand):
    data = tests.Data("examples", "alertmanager.json").json()

    def add_arguments(self, parser):
        parser.add_argument("--shard", default="Test Shard")
        parser.add_argument("--service", default=self.data["commonLabels"]["service"])
        parser.add_argument("--project", default=self.data["commonLabels"]["project"])

    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    def handle(self, shard, service, project, **kwargs):
        logging._handlers = []
        logging.basicConfig(level=logging.DEBUG)

        shard, _ = models.Shard.objects.get_or_create(name=shard)
        service, _ = models.Service.objects.get_or_create(name=service)
        project, _ = models.Project.objects.get_or_create(
            name=project, defaults={"shard": shard, "service": service}
        )

        alert = models.Alert.objects.create(body=json.dumps(self.data), error_count=1)

        tasks.process_alert(alert.pk)

        alert.alerterror_set.create(message="Test from CLI")



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from unittest import mock

from promgen import models, tests


class SignalTest(tests.PromgenTest):
    @mock.patch("promgen.models.Audit.log")
    @mock.patch("promgen.signals.trigger_write_config.send")
    def test_write_signal(self, write_mock, log_mock):
        # Create a test farm
        farm = models.Farm.objects.create(name="farm")
        models.Host.objects.create(name="Host", farm=farm)
        # Create a new project or testing
        project = models.Project.objects.create(name="Project", service_id=1, farm=farm, shard_id=1)
        e1 = models.Exporter.objects.create(
            job="Exporter 1",
            port=1234,
            project=project,
        )

        e2 = models.Exporter.objects.create(
            job="Exporter 2",
            port=1234,
            project=project,
        )

        # Should be called once for each created exporter
        self.assertEqual(write_mock.call_count, 2, "Two write calls")
        write_mock.assert_has_calls([mock.call(e1), mock.call(e2)])

    @mock.patch("promgen.models.Audit.log")
    @mock.patch("promgen.signals.trigger_write_config.send")
    def test_write_and_delete(self, write_mock, log_mock):
        # Create a test farm
        farm = models.Farm.objects.create(name="farm")
        models.Host.objects.create(name="Host", farm=farm)

        project = models.Project.objects.create(name="Project", service_id=1, farm=farm, shard_id=1)
        # Farm but no exporters so no call
        self.assertEqual(write_mock.call_count, 0, "Should not be called without exporters")

        models.Exporter.objects.create(
            job="Exporter 1",
            port=1234,
            project=project,
        )
        # Create an exporter so our call should be 1
        self.assertEqual(write_mock.call_count, 1, "Should be called after creating exporter")

        farm.delete()
        # Deleting our farm will call pre_delete on Farm and post_save on project
        self.assertEqual(write_mock.call_count, 3, "Should be called after deleting farm")

        models.Exporter.objects.create(
            job="Exporter 2",
            port=1234,
            project=project,
        )
        # Deleting our farm means our config is inactive, so no additional calls
        # from creating exporter
        self.assertEqual(
            write_mock.call_count,
            3,
            "No farms, so should not be called after deleting exporter",
        )



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import collections
import difflib
import json
from datetime import datetime
from urllib.parse import urlencode

import yaml
from pytz import timezone

from django import template
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

from promgen import util

register = template.Library()

EXCLUSION_MACRO = "<exclude>"


@register.filter()
def klass(value):
    return value.__class__.__name__


@register.filter()
def rule_dict(rule):
    return {
        "alert": rule.name,
        "expr": rulemacro(rule),
        "for": rule.duration,
        "labels": rule.labels,
        "annotations": rule.annotations,
    }


@register.filter()
def rulemacro(rule, clause=None):
    """
    Macro rule expansion

    Assuming a list of rules with children and parents, expand our macro to exclude child rules

    Can optionally pass expression to render in the context of the passed rule

    .. code-block:: none

        foo{<exclude>} / bar{<exclude>} > 5 # Parent Rule
        foo{project="A", <exclude>} / bar{project="A", <exclude>} > 3 # Child Rule
        foo{project="B"} / bar{project="B"} > 4 # Child Rule

        foo{project~="A|B"} / bar{project~="A|B"} > 5
        foo{project="A", } / bar{project="A"} > 3
        foo{project="B"} / bar{project="B"} > 4
    """

    if not clause:
        clause = rule.clause

    labels = collections.defaultdict(list)
    for r in rule.overrides.all():
        labels[r.content_type.model].append(r.content_object.name)

    filters = {k: "|".join(labels[k]) for k in sorted(labels)}
    macro = ",".join(sorted(f'{k}!~"{v}"' for k, v in filters.items()))
    return clause.replace(EXCLUSION_MACRO, macro)


@register.simple_tag
def diff_json(a, b):
    if isinstance(a, str):
        a = json.loads(a)
    if isinstance(b, str):
        b = json.loads(b)
    a = json.dumps(a, indent=4, sort_keys=True).splitlines(keepends=True)
    b = json.dumps(b, indent=4, sort_keys=True).splitlines(keepends=True)
    diff = "".join(difflib.unified_diff(a, b))
    if diff:
        return diff
    return "No Changes"


@register.filter()
def pretty_json(data):
    if isinstance(data, str):
        data = json.loads(data)
    return json.dumps(data, indent=4, sort_keys=True)


@register.filter()
def pretty_yaml(data):
    return yaml.safe_dump(data)


@register.filter()
def strftime(timestamp, fmt):
    tz = util.setting("timezone", "UTC")
    if isinstance(timestamp, int) or isinstance(timestamp, float):
        return timezone(tz).localize(datetime.fromtimestamp(timestamp)).strftime(fmt)
    return timestamp


@register.simple_tag
def breadcrumb(instance=None, label=None):
    """
    Create HTML Breadcrumb from instance

    Starting with the instance, walk up the tree building a bootstrap3
    compatible breadcrumb
    """
    from promgen import models

    def site(obj):
        yield reverse("site-detail"), obj.domain

    def shard(obj):
        yield reverse("datasource-list"), _("Datasource")
        yield obj.get_absolute_url(), obj.name

    def service(obj):
        yield reverse("service-list"), _("Services")
        yield obj.get_absolute_url(), obj.name

    def project(obj):
        yield from service(obj.service)
        yield obj.get_absolute_url(), obj.name

    def alert(obj):
        yield reverse("alert-list"), _("Alerts")
        yield obj.get_absolute_url(), obj.pk

    def rule(obj):
        if obj.content_type.model == "site":
            yield from site(obj.content_object)
        if obj.content_type.model == "service":
            yield from service(obj.content_object)
        if obj.content_type.model == "project":
            yield from project(obj.content_object)
        # If we have a new rule, it won't have a name
        if obj.pk:
            yield obj.get_absolute_url(), obj.name

    def sender(obj):
        if obj.content_type.model == "service":
            yield from service(obj.content_object)
        if obj.content_type.model == "project":
            yield from project(obj.content_object)

    def generator():
        yield reverse("home"), _("Home")
        if isinstance(instance, models.Sender):
            yield from sender(instance)
        if isinstance(instance, models.Project):
            yield from project(instance)
        if isinstance(instance, models.Service):
            yield from service(instance)
        if isinstance(instance, models.Shard):
            yield from shard(instance)
        if isinstance(instance, models.Rule):
            yield from rule(instance)
        if isinstance(instance, models.Alert):
            yield from alert(instance)

    def to_tag():
        yield '<ol class="breadcrumb">'
        for href, text in generator():
            yield format_html('<li><a href="{}">{}</a></li>', mark_safe(href), text)
        if label:
            yield format_html('<li class="active">{}</li>', _(label))
        yield "</ol>"

    return mark_safe("".join(to_tag()))


@register.simple_tag(takes_context=True)
def qs_replace(context, k, v):
    """
    Query string handler for paginators

    Assuming we have a query string like ?page=1&search=foo, there are several cases
    in which we want to replace only the page key, while leaving the rest alone.
    This tag allows us to replace individual values (like the current page) while
    carrying over other values (like a search string)

    Example:
    {% qs_replace 'page' page_obj.next_page_number %}
    """
    dict_ = context["request"].GET.copy()
    if v:
        dict_[k] = v
    else:
        dict_.pop(k, None)
    return dict_.urlencode()


@register.simple_tag
def urlqs(view, **kwargs):
    """
    Query string aware version of url template

    Instead of using {% url 'view' %}
    Use {% urlqs 'view' param=value %}

    This is useful for linking to pages that use filters.
    This only works for views that do not need additional parameters
    """
    return reverse(view) + "?" + urlencode(kwargs)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE
import collections
import datetime
import json
import logging
import subprocess
import tempfile
from urllib.parse import urljoin

import pytz
import yaml
from dateutil import parser

from django.core.exceptions import ValidationError
from django.utils import timezone

from promgen import models, renderers, serializers, util

logger = logging.getLogger(__name__)


def check_rules(rules):
    """
    Use promtool to check to see if a rule is valid or not

    The command name changed slightly from 1.x -> 2.x but this uses promtool
    to verify if the rules are correct or not. This can be bypassed by setting
    a dummy command such as /usr/bin/true that always returns true
    """

    with tempfile.NamedTemporaryFile(mode="w+b") as fp:
        logger.debug("Rendering to %s", fp.name)
        # Normally we wouldn't bother saving a copy to a variable here and would
        # leave it in the fp.write() call, but saving a copy in the variable
        # means we can see the rendered output in a Sentry stacktrace
        rendered = render_rules(rules)
        fp.write(rendered)
        fp.flush()

        # This command changed to be without a space in 2.x
        cmd = [util.setting("prometheus:promtool"), "check", "rules", fp.name]

        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT, encoding="utf8")
        except subprocess.CalledProcessError as e:
            raise ValidationError(message=e.output + rendered.decode("utf8"))


def render_rules(rules=None):
    """
    Render rules in a format that Prometheus understands

    :param rules: List of rules
    :type rules: list(Rule)
    :param int version: Prometheus rule format (1 or 2)
    :return: Returns rules in yaml or Prometheus v1 format
    :rtype: bytes

    This function can render in either v1 or v2 format
    We call prefetch_related_objects within this function to populate the
    other related objects that are mostly used for the sub lookups.
    """
    if rules is None:
        rules = models.Rule.objects.filter(enabled=True)

    return renderers.RuleRenderer().render(serializers.AlertRuleSerializer(rules, many=True).data)


def render_urls():
    urls = collections.defaultdict(list)

    for url in models.URL.objects.prefetch_related(
        "project__service",
        "project__shard",
        "project",
    ):
        urls[
            (
                url.project.name,
                url.project.service.name,
                url.project.shard.name,
                url.probe.module,
            )
        ].append(url.url)

    data = [
        {
            "labels": {
                "project": k[0],
                "service": k[1],
                "job": k[3],
                "__shard": k[2],
                "__param_module": k[3],
            },
            "targets": v,
        }
        for k, v in urls.items()
    ]
    return json.dumps(data, indent=2, sort_keys=True)


def render_config(service=None, project=None):
    data = []
    for exporter in models.Exporter.objects.prefetch_related(
        "project__farm__host_set",
        "project__farm",
        "project__service",
        "project__shard",
        "project",
    ):
        if not exporter.project.farm:
            continue
        if service and exporter.project.service.name != service.name:
            continue
        if project and exporter.project.name != project.name:
            continue
        if not exporter.enabled:
            continue

        labels = {
            "__shard": exporter.project.shard.name,
            "service": exporter.project.service.name,
            "project": exporter.project.name,
            "farm": exporter.project.farm.name,
            "__farm_source": exporter.project.farm.source,
            "job": exporter.job,
            "__scheme__": exporter.scheme,
        }
        if exporter.path:
            labels["__metrics_path__"] = exporter.path

        hosts = []
        for host in exporter.project.farm.host_set.all():
            hosts.append(f"{host.name}:{exporter.port}")

        data.append({"labels": labels, "targets": hosts})
    return json.dumps(data, indent=2, sort_keys=True)


def import_rules_v2(config, content_object=None):
    """
    Loop through a dictionary and add rules to the database

    This assumes a dictionary in the 2.x rule format.
    See promgen/tests/examples/import.rule.yml for an example
    """
    # If not already a dictionary, try to load as YAML
    if not isinstance(config, dict):
        config = yaml.safe_load(config)

    # If 'groups' does not exist in our config, assume that a single
    # rule is being imported (perhaps from Promgen's UI) so wrap it
    # to be the full rule format we expect
    # https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/
    # TODO In the future want to refactor this into the API itself
    if "groups" not in config:
        config = {"groups": [{"name": "Import", "rules": [config]}]}

    counters = collections.defaultdict(int)
    for group in config["groups"]:
        for r in group["rules"]:
            defaults = {
                "clause": r["expr"],
                "duration": r["for"],
                "labels": r.get("labels", {}),
                "annotations": r.get("annotations", {}),
            }

            # Check our labels to see if we have a project or service
            # label set and if not, default it to a global rule
            if content_object:
                defaults["obj"] = content_object
            elif "project" in defaults["labels"]:
                defaults["obj"] = models.Project.objects.get(name=defaults["labels"]["project"])
            elif "service" in defaults["labels"]:
                defaults["obj"] = models.Service.objects.get(name=defaults["labels"]["service"])
            else:
                defaults["obj"] = models.Site.objects.get_current()

            _, created = models.Rule.objects.get_or_create(name=r["alert"], defaults=defaults)

            if created:
                counters["Rules"] += 1

    return dict(counters)


def import_config(config, replace_shard=None):
    counters = collections.defaultdict(list)
    skipped = collections.defaultdict(list)
    for entry in config:
        if replace_shard:
            logger.debug("Importing into shard %s", replace_shard)
            entry["labels"]["__shard"] = replace_shard
        shard, created = models.Shard.objects.get_or_create(
            name=entry["labels"].get("__shard", "Default")
        )
        if created:
            logger.debug("Created shard %s", shard)
            counters["Shard"].append(shard)
        else:
            skipped["Shard"].append(shard)

        service, created = models.Service.objects.get_or_create(
            name=entry["labels"]["service"],
        )
        if created:
            logger.debug("Created service %s", service)
            counters["Service"].append(service)
        else:
            skipped["Service"].append(service)

        farm, created = models.Farm.objects.get_or_create(
            name=entry["labels"]["farm"],
            defaults={"source": entry["labels"].get("__farm_source", "pmc")},
        )
        if created:
            logger.debug("Created farm %s", farm)
            counters["Farm"].append(farm)
        else:
            skipped["Farm"].append(farm)

        project, created = models.Project.objects.get_or_create(
            name=entry["labels"]["project"],
            service=service,
            shard=shard,
            defaults={"farm": farm},
        )
        if created:
            logger.debug("Created project %s", project)
            counters["Project"].append(project)
        elif project.farm != farm:
            logger.debug("Linking farm [%s] with [%s]", farm, project)
            project.farm = farm
            project.save()

        for target in entry["targets"]:
            target, port = target.split(":")
            host, created = models.Host.objects.get_or_create(
                name=target,
                farm_id=farm.id,
            )

            if created:
                logger.debug("Created host %s", host)
                counters["Host"].append(host)

            exporter, created = models.Exporter.objects.get_or_create(
                job=entry["labels"]["job"],
                port=port,
                project=project,
                path=entry["labels"].get("__metrics_path__", ""),
            )

            if created:
                logger.debug("Created exporter %s", exporter)
                counters["Exporter"].append(exporter)

    return counters, skipped


def silence(*, labels, duration=None, **kwargs):
    """
    Post a silence message to Alert Manager
    Duration should be sent in a format like 1m 2h 1d etc
    """

    if duration:
        start = timezone.now()
        if duration.endswith("m"):
            end = start + datetime.timedelta(minutes=int(duration[:-1]))
        elif duration.endswith("h"):
            end = start + datetime.timedelta(hours=int(duration[:-1]))
        elif duration.endswith("d"):
            end = start + datetime.timedelta(days=int(duration[:-1]))
        else:
            raise ValidationError("Unknown time modifier")
        kwargs["startsAt"] = start.isoformat()
        kwargs["endsAt"] = end.isoformat()
    else:
        local_timezone = pytz.timezone(util.setting("timezone", "UTC"))
        for key in ["startsAt", "endsAt"]:
            kwargs[key] = local_timezone.localize(parser.parse(kwargs[key])).isoformat()

    kwargs["matchers"] = [
        {
            "name": name,
            "value": value,
            "isEqual": True,  # Right now we only support =~
            "isRegex": True if value.endswith("*") else False,
        }
        for name, value in labels.items()
    ]

    logger.debug("Sending silence for %s", kwargs)
    url = urljoin(util.setting("alertmanager:url"), "/api/v2/silences")
    response = util.post(url, json=kwargs)
    response.raise_for_status()
    return response



# Generated by Django 2.2.10 on 2020-02-14 05:41


import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0012_validation_updates'),
    ]

    operations = [
        migrations.AlterField(
            model_name='farm',
            name='name',
            field=models.CharField(max_length=128, validators=[django.core.validators.RegexValidator('^[\\w][- \\w]+\\Z', 'Unicode letters, numbers, underscores, or hyphens or spaces')]),
        ),
        migrations.AlterField(
            model_name='project',
            name='name',
            field=models.CharField(max_length=128, unique=True, validators=[django.core.validators.RegexValidator('^[\\w][- \\w]+\\Z', 'Unicode letters, numbers, underscores, or hyphens or spaces')]),
        ),
        migrations.AlterField(
            model_name='rule',
            name='name',
            field=models.CharField(max_length=128, unique=True, validators=[django.core.validators.RegexValidator('[a-zA-Z_:][a-zA-Z0-9_:]*', 'Only alphanumeric characters are allowed.')]),
        ),
        migrations.AlterField(
            model_name='service',
            name='name',
            field=models.CharField(max_length=128, unique=True, validators=[django.core.validators.RegexValidator('^[\\w][- \\w]+\\Z', 'Unicode letters, numbers, underscores, or hyphens or spaces')]),
        ),
        migrations.AlterField(
            model_name='shard',
            name='name',
            field=models.CharField(max_length=128, unique=True, validators=[django.core.validators.RegexValidator('^[\\w][- \\w]+\\Z', 'Unicode letters, numbers, underscores, or hyphens or spaces')]),
        ),
    ]



# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from django.core.management.base import BaseCommand

from promgen import models, util


class Command(BaseCommand):
    def add_arguments(self, parser):
        help_text = util.help_text(models.Exporter)

        parser.add_argument(
            "project",
            type=util.cast(models.Project),
            help="Existing Project",
        )
        parser.add_argument("job", help=help_text("job"))
        parser.add_argument("port", type=int, help=help_text("port"))
        parser.add_argument("path", default="", nargs="?", help=help_text("path"))
        parser.add_argument(
            "--enabled",
            default=False,
            action="store_true",
            help=help_text("enabled"),
        )

    def handle(self, project, **kwargs):
        job, created = models.Exporter.objects.get_or_create(
            project=project,
            job=kwargs["job"],
            port=kwargs["port"],
            defaults={"path": kwargs["path"], "enabled": kwargs["enabled"]},
        )
        if created:
            print("Registered job", job)
        else:
            print("Found existing job", job)



import collections

from rest_framework import serializers

from django.db.models import prefetch_related_objects

import promgen.templatetags.promgen as macro
from promgen import models, shortcuts
from promgen.shortcuts import resolve_domain


class WebLinkField(serializers.Field):
    def get_attribute(self, instance):
        return instance

    def to_representation(self, obj):
        return shortcuts.resolve_domain(obj.get_absolute_url())


class ShardSerializer(serializers.ModelSerializer):
    html = WebLinkField()

    class Meta:
        model = models.Shard
        exclude = ("id",)
        lookup_field = "name"


class ServiceSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source="owner.username")
    shard = serializers.ReadOnlyField(source="shard.name")
    html = WebLinkField()

    class Meta:
        model = models.Service
        exclude = ("id",)
        lookup_field = "name"


class ProjectSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source="owner.username")
    service = serializers.ReadOnlyField(source="service.name")
    shard = serializers.ReadOnlyField(source="service.shard.name")
    html = WebLinkField()

    class Meta:
        model = models.Project
        lookup_field = "name"
        exclude = ("id", "farm")


class SenderSerializer(serializers.ModelSerializer):
    owner = serializers.ReadOnlyField(source="owner.username")
    label = serializers.ReadOnlyField(source="show_value")

    class Meta:
        model = models.Sender


class AlertRuleList(serializers.ListSerializer):
    def to_representation(self, data):
        grouped_list = collections.defaultdict(list)
        for item in data:
            data = self.child.to_representation(item)
            grouped_list[str(item.content_object)].append(data)
        return grouped_list

    @property
    def data(self):
        if not hasattr(self, "_data"):
            self._data = self.to_representation(self.instance)
        return serializers.ReturnDict(self._data, serializer=self)


class AlertRuleSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Rule
        fields = ("alert", "expr", "for", "labels", "annotations")

    @classmethod
    def many_init(cls, queryset, *args, **kwargs):
        # when rendering many items at once, we want to make sure we
        # do our prefetch operations in one go, before passing it off
        # to our custom list renderer to group things in a dictionary
        kwargs["child"] = cls()
        prefetch_related_objects(
            queryset,
            "content_object",
            "content_type",
            "overrides__content_object",
            "overrides__content_type",
        )
        return AlertRuleList(queryset, *args, **kwargs)

    def to_representation(self, obj):
        annotations = obj.annotations
        annotations["rule"] = resolve_domain("rule-detail", pk=obj.pk if obj.pk else 0)

        return {
            "alert": obj.name,
            "expr": macro.rulemacro(obj),
            "for": obj.duration,
            "labels": obj.labels,
            "annotations": annotations,
        }



# Generated by Django 2.1.2 on 2018-11-28 08:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('promgen', '0005_project_owner'),
    ]

    operations = [
        migrations.CreateModel(
            name='DefaultExporter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('job', models.CharField(max_length=128)),
                ('port', models.IntegerField()),
                ('path', models.CharField(blank=True, max_length=128)),
            ],
            options={
                'ordering': ['job', 'port'],
            },
        ),
        migrations.AlterUniqueTogether(
            name='defaultexporter',
            unique_together={('job', 'port', 'path')},
        ),
    ]



# Copyright (c) 2019 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from unittest import mock

from django.core import management
from django.core.management.base import CommandError

from promgen import models
from promgen.tests import PromgenTest


class CLITests(PromgenTest):
    fixtures = ["testcases.yaml", "extras.yaml"]

    @mock.patch("promgen.signals._trigger_write_config")
    def test_register_job(self, mock_signal):
        # Assert when project doesn't exist
        with self.assertRaises(CommandError):
            management.call_command("register-job", "missing-project", "example", 1234)

        management.call_command("register-job", "test-project", "example", 1234)

        # Ensure the jobs we expect exist
        self.assertCount(models.Exporter, 1)

        # Registering the same job again shouldn't change our count
        management.call_command("register-job", "test-project", "example", 1234)
        self.assertCount(models.Exporter, 1)

        # But registering a new one will
        management.call_command("register-job", "test-project", "example", 4321)
        self.assertCount(models.Exporter, 2)

    @mock.patch("promgen.signals._trigger_write_config")
    def test_register_host(self, mock_signal):
        # Assert when project doesn't exist
        with self.assertRaises(CommandError):
            management.call_command("register-host", "missing-project", "cli.example.com")

        project = models.Project.objects.create(name="cli-project", service_id=1, shard_id=1)

        # Still assert an error if there is no Farm
        with self.assertRaises(CommandError):
            management.call_command("register-host", "cli-project", "cli.example.com")

        # Register farm and finally register host
        project.farm = models.Farm.objects.create(name="cli-farm")
        project.save()
        management.call_command("register-host", "cli-project", "cli.example.com")
        self.assertCount(models.Host, 2, "Should be a new host registered (1 host from fixture)")



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE


from django.urls import reverse

from promgen import models, validators
from promgen.tests import PromgenTest


class HostTests(PromgenTest):
    def setUp(self):
        self.force_login(username="demo")

    # For our first two tests, we just want to make sure that both newline
    # separated and comma separated work, but are not necessarily testing
    # valid/invalid hostnames
    def test_newline(self):
        self.client.post(
            reverse("hosts-add", args=[1]),
            {"hosts": "\naaa.example.com\nbbb.example.com\nccc.example.com \n"},
            follow=False,
        )
        self.assertCount(models.Host, 3, "Expected 3 hosts")

    def test_comma(self):
        self.client.post(
            reverse("hosts-add", args=[1]),
            {"hosts": ",,aaa.example.com, bbb.example.com,ccc.example.com,"},
            follow=False,
        )
        self.assertCount(models.Host, 3, "Expected 3 hosts")

    # Within our new host code, the hosts are split (by newline or comma) and then
    # individually tested. Here we will test our validator on specific entries that
    # should pass or fail
    def test_validators(self):
        # Hostname only should be valid
        validators.hostname("bare-hostname")
        # FQDN should be valid
        validators.hostname("fqdn.example.com")
        # UPPERCASE and trailing dot should also be fine
        validators.hostname("FQDN.with.trailing.dot.example.com.")
        # Hostname cannot contain underscore
        with self.assertRaises(validators.ValidationError):
            validators.hostname("underscore_in_hostname")
        # Hostname should not include port
        with self.assertRaises(validators.ValidationError):
            validators.hostname("invalid:host")
        # Hostname should not be a url (no scheme)
        with self.assertRaises(validators.ValidationError):
            validators.hostname("http://example.com")
        # Hostnames should not contain a path component
        with self.assertRaises(validators.ValidationError):
            validators.hostname("example.com/path")



# Copyright (c) 2018 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE


from django.test import override_settings

from promgen import models, rest, tests


class RestAPITest(tests.PromgenTest):
    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    def test_alert_blackhole(self):
        response = self.fireAlert("heartbeat.json")
        self.assertRoute(response, rest.AlertReceiver, 202)
        self.assertCount(models.Alert, 0, "Heartbeat alert should be deleted")

    @override_settings(PROMGEN=tests.SETTINGS)
    @override_settings(CELERY_TASK_ALWAYS_EAGER=True)
    def test_alert(self):
        response = self.fireAlert()
        self.assertEqual(response.status_code, 202)
        self.assertCount(models.Alert, 1, "Alert Queued")



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from django.core.management.base import BaseCommand

from promgen.models import Prometheus, Shard


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("shard")
        parser.add_argument("host")
        parser.add_argument("port", type=int)

    def handle(self, shard, host, port, **kwargs):
        shard, created = Shard.objects.get_or_create(name=shard)
        if created:
            self.stdout.write("Created shard " + shard.name)

        server, created = Prometheus.objects.get_or_create(
            host=host, port=port, defaults={"shard": shard}
        )
        if created:
            self.stdout.write(f"Created {server} on {shard.name}")
        else:
            old_shard = server.shard
            server.shard = shard
            server.save()
            self.stdout.write(f"Moved {server} from {old_shard.name} to {shard.name}")



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

"""
Django settings for promgen project.

Generated by 'django-admin startproject' using Django 1.10.2.

For more information on this file, see
https://docs.djangoproject.com/en/1.10/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.10/ref/settings/
"""

import importlib.metadata
import os
import pathlib

import environ
import yaml

from django.urls import reverse_lazy

from promgen import PROMGEN_CONFIG_FILE
from promgen.plugins import apps_from_setuptools

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = pathlib.Path(__file__).parent.parent
DOT_ENV = BASE_DIR / ".env"
PROMGEN_VERSION = importlib.metadata.version("promgen")

# Load our environment
env = environ.Env()
if DOT_ENV.exists():
    env.read_env(DOT_ENV)

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = env.str("SECRET_KEY")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = env.bool("DEBUG", default=False)

if DEBUG:
    INTERNAL_IPS = ["127.0.0.1"]

# Settings for Prometheus paths and such
if PROMGEN_CONFIG_FILE.exists():
    with PROMGEN_CONFIG_FILE.open() as fp:
        PROMGEN = yaml.safe_load(fp)
else:
    PROMGEN = {}

PROMGEN_DEFAULT_GROUP = "Default"
PROMGEN_SCHEME = env.str("PROMGEN_SCHEME", default="http")

ALLOWED_HOSTS = env.list("ALLOWED_HOSTS", default=["127.0.0.1", "localhost"])


# Application definition

INSTALLED_APPS = apps_from_setuptools + [
    "promgen",
    # Third Party
    "django_filters",
    "rest_framework.authtoken",
    "rest_framework",
    "social_django",
    # Django
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.sites",
    "django.contrib.staticfiles",
]

# We explicitly include debug_toolbar and whitenoise here, but selectively
# remove it below, so that we can more easily control the import order
MIDDLEWARE = [
    "debug_toolbar.middleware.DebugToolbarMiddleware",  # Only enabled for debug
    "django.middleware.security.SecurityMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",  # Used primarily for docker
    "django.middleware.locale.LocaleMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "promgen.middleware.PromgenMiddleware",
]

SOCIAL_AUTH_RAISE_EXCEPTIONS = DEBUG
LOGIN_URL = reverse_lazy("login")
LOGIN_REDIRECT_URL = reverse_lazy("home")
LOGOUT_REDIRECT_URL = reverse_lazy("home")

ROOT_URLCONF = "promgen.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "promgen.context_processors.settings_in_view",
                "social_django.context_processors.backends",
                "social_django.context_processors.login_redirect",
            ],
        },
    },
]

WSGI_APPLICATION = "promgen.wsgi.application"


# Database
# https://docs.djangoproject.com/en/1.10/ref/settings/#databases

DATABASES = {
    "default": env.db("DATABASE_URL", default="sqlite:///" + str(BASE_DIR / "db.sqlite3")),
}


# Password validation
# https://docs.djangoproject.com/en/1.10/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
# https://docs.djangoproject.com/en/1.10/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"
STATIC_ROOT_DEFAULT = pathlib.Path.home() / ".cache" / "promgen"
STATIC_ROOT = env.str("STATIC_ROOT", default=str(STATIC_ROOT_DEFAULT))

try:
    # If whitenoise is not available, we will remove it from our middleware
    import whitenoise.middleware  # NOQA
except ImportError:
    STATICFILES_STORAGE = "django.contrib.staticfiles.storage.ManifestStaticFilesStorage"
    MIDDLEWARE.remove("whitenoise.middleware.WhiteNoiseMiddleware")
else:
    STATICFILES_STORAGE = "whitenoise.storage.CompressedManifestStaticFilesStorage"


SITE_ID = 1

if "SENTRY_DSN" in os.environ:
    import sentry_sdk
    from sentry_sdk.integrations.django import DjangoIntegration
    from sentry_sdk.integrations.celery import CeleryIntegration

    os.environ.setdefault("SENTRY_RELEASE", PROMGEN_VERSION)
    # By default we want to keep sentry's default of NOT sending user
    # information, but in some environments, it's useful for debugging
    # purposes so lets add a flag here that we can use.
    sentry_sdk.init(
        integrations=[DjangoIntegration(), CeleryIntegration()],
        send_default_pii=env.bool("SENTRY_SEND_DEFAULT_PII", False),
    )


REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.TokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": (
        "rest_framework.permissions.DjangoModelPermissionsOrAnonReadOnly",
    ),
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
}

# If CELERY_BROKER_URL is set in our environment, then we configure celery as
# expected. If it is not configured, then we set CELERY_TASK_ALWAYS_EAGER to
# force celery to run all tasks in the same process (effectively runs each task
# as a normal function)
try:
    CELERY_BROKER_URL = env.str("CELERY_BROKER_URL")
except Exception:
    CELERY_TASK_ALWAYS_EAGER = True


try:
    # If debug_toolbar is not available, we will remove it from our middleware
    import debug_toolbar  # NOQA

    INSTALLED_APPS += ["debug_toolbar"]
except ImportError:
    MIDDLEWARE.remove("debug_toolbar.middleware.DebugToolbarMiddleware")


# Load overrides from PROMGEN to replace Django settings
for k, v in PROMGEN.pop("django", {}).items():
    globals()[k] = v

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import collections
import concurrent.futures
import datetime
import json
import logging
import platform
import time
from itertools import chain

import prometheus_client
import requests
from prometheus_client.core import CounterMetricFamily, GaugeMetricFamily

import promgen.templatetags.promgen as macro
from promgen import (
    discovery,
    forms,
    mixins,
    models,
    plugins,
    prometheus,
    signals,
    tasks,
    util,
)
from promgen.shortcuts import resolve_domain

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, Q
from django.db.utils import IntegrityError
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import ugettext as _
from django.views.generic import DetailView, ListView, UpdateView, View
from django.views.generic.base import RedirectView, TemplateView
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import CreateView, DeleteView, FormView

logger = logging.getLogger(__name__)


class DatasourceList(LoginRequiredMixin, ListView):
    queryset = models.Shard.objects.prefetch_related(
        "project_set__service",
        "project_set__service__owner",
        "project_set__service__notifiers",
        "project_set__service__notifiers__owner",
        "project_set__service__rule_set",
        "project_set",
        "project_set__owner",
        "project_set__farm",
        "project_set__exporter_set",
        "project_set__notifiers",
        "project_set__notifiers__owner",
        "prometheus_set",
    ).annotate(num_projects=Count("project"))


class DatasourceDetail(LoginRequiredMixin, DetailView):
    queryset = models.Shard.objects.prefetch_related(
        "project_set__service",
        "project_set__service__owner",
        "project_set__service__notifiers",
        "project_set__service__notifiers__owner",
        "project_set__service__notifiers__filter_set",
        "project_set__service__rule_set",
        "project_set",
        "project_set__owner",
        "project_set__farm",
        "project_set__exporter_set",
        "project_set__notifiers",
        "project_set__notifiers__owner",
        "project_set__notifiers__filter_set",
    )


class ServiceList(LoginRequiredMixin, ListView):
    paginate_by = 20
    queryset = models.Service.objects.prefetch_related(
        "rule_set",
        "rule_set__parent",
        "project_set",
        "project_set__owner",
        "project_set__shard",
        "project_set__notifiers",
        "project_set__notifiers__owner",
        "project_set__notifiers__filter_set",
        "project_set__farm",
        "project_set__exporter_set",
        "owner",
        "notifiers",
        "notifiers__owner",
        "notifiers__filter_set",
    )


class HomeList(LoginRequiredMixin, ListView):
    template_name = "promgen/home.html"

    def get_queryset(self):
        # TODO: Support showing subscribed projects as well
        # Get the list of senders that a user is currently subscribed to
        senders = models.Sender.objects.filter(
            value=self.request.user.username,
            sender="promgen.notification.user",
            content_type=ContentType.objects.get_for_model(models.Service),
        ).values_list("object_id")

        # and return just our list of services
        return models.Service.objects.filter(pk__in=senders).prefetch_related(
            "notifiers",
            "notifiers__owner",
            "owner",
            "rule_set",
            "rule_set__parent",
            "project_set",
            "project_set__farm",
            "project_set__shard",
            "project_set__exporter_set",
            "project_set__notifiers",
            "project_set__owner",
            "project_set__notifiers__owner",
        )


class HostList(LoginRequiredMixin, ListView):
    queryset = models.Host.objects.prefetch_related(
        "farm",
        "farm__project_set",
        "farm__project_set__service",
    )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["host_groups"] = collections.defaultdict(list)
        for host in context["object_list"]:
            context["host_groups"][host.name].append(host)
        context["host_groups"] = dict(context["host_groups"])
        return context


class HostDetail(LoginRequiredMixin, View):
    def get(self, request, slug):
        context = {}
        context["slug"] = self.kwargs["slug"]

        context["host_list"] = models.Host.objects.filter(
            name__icontains=self.kwargs["slug"]
        ).prefetch_related("farm")

        if not context["host_list"]:
            return render(request, "promgen/host_404.html", context, status=404)

        context["farm_list"] = models.Farm.objects.filter(
            id__in=context["host_list"].values_list("farm_id", flat=True)
        )

        context["project_list"] = models.Project.objects.filter(
            id__in=context["farm_list"].values_list("project__id", flat=True)
        ).prefetch_related("notifiers", "rule_set")

        context["exporter_list"] = models.Exporter.objects.filter(
            project_id__in=context["project_list"].values_list("id", flat=True)
        ).prefetch_related("project", "project__service")

        context["service_list"] = models.Service.objects.filter(
            id__in=context["project_list"].values_list("service__id", flat=True)
        ).prefetch_related("notifiers", "rule_set")

        context["rule_list"] = (
            models.Rule.objects.filter(
                Q(id__in=context["project_list"].values_list("rule_set__id"))
                | Q(id__in=context["service_list"].values_list("rule_set__id"))
                | Q(id__in=models.Site.objects.get_current().rule_set.values_list("id"))
            )
            .select_related("content_type")
            .prefetch_related("content_object")
        )

        context["notifier_list"] = (
            models.Sender.objects.filter(
                Q(id__in=context["project_list"].values_list("notifiers__id"))
                | Q(id__in=context["service_list"].values_list("notifiers__id"))
            )
            .select_related("content_type")
            .prefetch_related("content_object")
        )

        return render(request, "promgen/host_detail.html", context)


class AuditList(LoginRequiredMixin, ListView):
    model = models.Audit

    FILTERS = {
        "project": models.Project,
        "service": models.Service,
        "rule": models.Rule,
    }

    def get_queryset(self):
        queryset = self.model.objects.order_by("-created").prefetch_related(
            "content_object", "user"
        )

        for key in self.FILTERS:
            if key in self.request.GET:
                obj = self.FILTERS[key].objects.get(pk=self.request.GET[key])
                # Get any log entries for the object itself
                qset = Q(
                    object_id=obj.id,
                    content_type_id=ContentType.objects.get_for_model(obj).id,
                )
                if key in ["project", "service"]:
                    # Look for any registered notifiers
                    qset |= Q(
                        content_type_id=ContentType.objects.get_for_model(models.Sender).id,
                        object_id__in=obj.notifiers.values_list("id", flat=True),
                    )
                    # Look for any registered rules
                    qset |= Q(
                        content_type_id=ContentType.objects.get_for_model(models.Rule).id,
                        object_id__in=obj.rule_set.values_list("id", flat=True),
                    )
                if key == "service":
                    # Only services may have projects
                    qset |= Q(
                        content_type_id=ContentType.objects.get_for_model(models.Project).id,
                        object_id__in=obj.project_set.values_list("id", flat=True),
                    )
                if key == "project":
                    # Only projects may have exporters
                    qset |= Q(
                        content_type_id=ContentType.objects.get_for_model(models.Exporter).id,
                        object_id__in=obj.exporter_set.values_list("id", flat=True),
                    )
                    # Only projects may have URLs
                    qset |= Q(
                        content_type_id=ContentType.objects.get_for_model(models.URL).id,
                        object_id__in=obj.url_set.values_list("id", flat=True),
                    )
                queryset = queryset.filter(qset)
        if "user" in self.request.GET:
            queryset = queryset.filter(user_id=self.request.GET["user"])

        return queryset

    paginate_by = 50


class ServiceDetail(LoginRequiredMixin, DetailView):
    queryset = models.Service.objects.prefetch_related(
        "rule_set",
        "notifiers",
        "notifiers__filter_set",
        "notifiers__owner",
        "project_set",
        "project_set__shard",
        "project_set__farm",
        "project_set__exporter_set",
        "project_set__notifiers",
        "project_set__notifiers__owner",
    )


class ServiceDelete(LoginRequiredMixin, DeleteView):
    model = models.Service

    def get_success_url(self):
        return reverse("service-list")


class ProjectDelete(LoginRequiredMixin, DeleteView):
    model = models.Project

    def get_success_url(self):
        return reverse("service-detail", args=[self.object.service_id])


class NotifierUpdate(LoginRequiredMixin, UpdateView):
    model = models.Sender
    form_class = forms.NotifierUpdate

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        obj = self.get_object()
        # For populating breadcrumb
        context[obj.content_type.model] = obj.content_object
        return context

    def post(self, request, pk):
        if "filter.pk" in request.POST:
            f = models.Filter.objects.get(pk=request.POST["filter.pk"])
            f.delete()
            messages.success(request, f"Removed filter {f.name} {f.value}")
        if "filter.name" in request.POST:
            obj = self.get_object()
            f, created = obj.filter_set.get_or_create(
                name=request.POST["filter.name"], value=request.POST["filter.value"]
            )
            if created:
                messages.success(request, f"Created filter {f.name} {f.value}")
            else:
                messages.warning(request, f"Updated filter {f.name} {f.value}")
        if "next" in request.POST:
            return redirect(request.POST["next"])
        return self.get(self, request, pk)


class NotifierDelete(LoginRequiredMixin, DeleteView):
    model = models.Sender

    def get_success_url(self):
        if "next" in self.request.POST:
            return self.request.POST["next"]
        if hasattr(self.object.content_object, "get_absolute_url"):
            return self.object.content_object.get_absolute_url()
        return reverse("profile")


class NotifierTest(LoginRequiredMixin, View):
    def post(self, request, pk):
        sender = get_object_or_404(models.Sender, id=pk)
        try:
            sender.test()
        except Exception:
            messages.warning(request, "Error sending test message with " + sender.sender)
        else:
            messages.info(request, "Sent test message with " + sender.sender)

        if "next" in request.POST:
            return redirect(request.POST["next"])
        if hasattr(sender.content_object, "get_absolute_url"):
            return redirect(sender.content_object)
        return redirect("profile")


class ExporterDelete(LoginRequiredMixin, DeleteView):
    model = models.Exporter

    def get_success_url(self):
        return reverse("project-detail", args=[self.object.project_id])


class ExporterToggle(LoginRequiredMixin, View):
    def post(self, request, pk):
        exporter = get_object_or_404(models.Exporter, id=pk)
        exporter.enabled = not exporter.enabled
        exporter.save()
        signals.trigger_write_config.send(request)
        return JsonResponse({"redirect": exporter.project.get_absolute_url()})


class NotifierToggle(LoginRequiredMixin, View):
    def post(self, request, pk):
        sender = get_object_or_404(models.Sender, id=pk)
        sender.enabled = not sender.enabled
        sender.save()
        # Redirect to current page
        return JsonResponse({"redirect": ""})


class RuleDelete(mixins.PromgenPermissionMixin, DeleteView):
    model = models.Rule

    def get_permission_denied_message(self):
        return "Unable to delete rule %s. User lacks permission" % self.object

    def get_permission_required(self):
        # In the case of rules, we want to make sure the user has permission
        # to delete the rule itself, but also permission to change the linked object
        self.object = self.get_object()
        obj = self.object._meta
        tgt = self.object.content_object._meta

        yield f"{obj.app_label}.delete_{obj.model_name}"
        yield f"{tgt.app_label}.change_{tgt.model_name}"

    def get_success_url(self):
        return self.object.content_object.get_absolute_url()


class RuleToggle(mixins.PromgenPermissionMixin, SingleObjectMixin, View):
    model = models.Rule

    def get_permission_denied_message(self):
        return "Unable to toggle rule %s. User lacks permission" % self.object

    def get_permission_required(self):
        # In the case of rules, we want to make sure the user has permission
        # to delete the rule itself, but also permission to change the linked object
        self.object = self.get_object()
        obj = self.object._meta
        tgt = self.object.content_object._meta

        yield f"{obj.app_label}.change_{obj.model_name}"
        yield f"{tgt.app_label}.change_{tgt.model_name}"

    def post(self, request, pk):
        self.object.enabled = not self.object.enabled
        self.object.save()
        return JsonResponse({"redirect": self.object.content_object.get_absolute_url()})


class HostDelete(LoginRequiredMixin, DeleteView):
    model = models.Host

    def get_success_url(self):
        # If there's only one linked project then we redirect to the project page
        # otherwise we redirect to our farm page
        if self.object.farm.project_set.count():
            return self.object.farm.project_set.first().get_absolute_url()
        return self.object.farm.get_absolute_url()


class ProjectDetail(LoginRequiredMixin, DetailView):
    queryset = models.Project.objects.prefetch_related(
        "rule_set",
        "rule_set__parent",
        "notifiers",
        "notifiers__owner",
        "shard",
        "service",
        "service__rule_set",
        "service__rule_set__parent",
    )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["sources"] = models.Farm.driver_set()
        context["url_form"] = forms.URLForm()
        return context


class FarmList(LoginRequiredMixin, ListView):
    paginate_by = 50
    queryset = models.Farm.objects.prefetch_related(
        "project_set",
        "host_set",
    )


class FarmDetail(LoginRequiredMixin, DetailView):
    model = models.Farm


class FarmUpdate(LoginRequiredMixin, UpdateView):
    model = models.Farm
    button_label = _("Update Farm")
    template_name = "promgen/farm_form.html"
    form_class = forms.FarmForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["project"] = self.object.project_set.first()
        context["service"] = context["project"].service
        return context

    def form_valid(self, form):
        farm, created = models.Farm.objects.update_or_create(
            id=self.kwargs["pk"],
            defaults=form.clean(),
        )
        return HttpResponseRedirect(reverse("project-detail", args=[farm.project_set.first().id]))


class FarmDelete(LoginRequiredMixin, RedirectView):
    pattern_name = "farm-detail"

    def post(self, request, pk):
        farm = get_object_or_404(models.Farm, id=pk)
        farm.delete()

        return HttpResponseRedirect(request.POST.get("next", reverse("service-list")))


class UnlinkFarm(LoginRequiredMixin, View):
    def post(self, request, pk):
        project = get_object_or_404(models.Project, id=pk)
        oldfarm, project.farm = project.farm, None
        project.save()
        signals.trigger_write_config.send(request)

        if oldfarm.project_set.count() == 0 and oldfarm.editable is False:
            logger.debug("Cleaning up old farm %s", oldfarm)
            oldfarm.delete()

        return HttpResponseRedirect(reverse("project-detail", args=[project.id]))


class RulesList(LoginRequiredMixin, ListView, mixins.ServiceMixin):
    template_name = "promgen/rule_list.html"
    queryset = models.Rule.objects.prefetch_related("content_type", "content_object")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        site_rules = models.Rule.objects.filter(
            content_type__model="site", content_type__app_label="promgen"
        ).prefetch_related("content_object")

        service_rules = models.Rule.objects.filter(
            content_type__model="service", content_type__app_label="promgen"
        ).prefetch_related(
            "content_object",
            "content_object",
            "parent",
        )

        project_rules = models.Rule.objects.filter(
            content_type__model="project", content_type__app_label="promgen"
        ).prefetch_related(
            "content_object",
            "content_object__service",
            "content_object__service",
            "parent",
        )

        context["rule_list"] = chain(site_rules, service_rules, project_rules)

        return context


class RulesCopy(LoginRequiredMixin, View):
    def post(self, request, pk):
        original = get_object_or_404(models.Rule, id=pk)
        form = forms.RuleCopyForm(request.POST)

        if form.is_valid():
            rule = original.copy_to(**form.clean())
            return HttpResponseRedirect(reverse("rule-edit", args=[rule.id]))
        else:
            return HttpResponseRedirect(reverse("service-detail", args=[pk]))


class FarmRefresh(LoginRequiredMixin, RedirectView):
    pattern_name = "farm-detail"

    def post(self, request, pk):
        farm = get_object_or_404(models.Farm, id=pk)
        # If any hosts are added or removed, then we want to
        # trigger a config refresh
        if any(farm.refresh()):
            signals.trigger_write_config.send(request)
        messages.info(request, "Refreshed hosts")
        if "next" in request.POST:
            return HttpResponseRedirect(request.POST["next"])
        # If we don't have an explicit redirect, we can redirect to the farm
        # itself
        return redirect(farm)


class FarmConvert(LoginRequiredMixin, RedirectView):
    pattern_name = "farm-detail"

    def post(self, request, pk):
        farm = get_object_or_404(models.Farm, id=pk)
        farm.source = discovery.FARM_DEFAULT

        try:
            farm.save()
        except IntegrityError:
            return render(
                request,
                "promgen/farm_duplicate.html",
                {
                    "pk": farm.pk,
                    "next": request.POST.get("next", reverse("farm-detail", args=[farm.pk])),
                    "farm_list": models.Farm.objects.filter(name=farm.name),
                },
            )

        return HttpResponseRedirect(
            request.POST.get("next", reverse("farm-detail", args=[farm.pk]))
        )


class FarmLink(LoginRequiredMixin, View):
    def get(self, request, pk, source):
        context = {
            "source": source,
            "project": get_object_or_404(models.Project, id=pk),
            "farm_list": sorted(models.Farm.fetch(source=source)),
        }
        return render(request, "promgen/link_farm.html", context)

    def post(self, request, pk, source):
        project = get_object_or_404(models.Project, id=pk)
        farm, created = models.Farm.objects.get_or_create(
            name=request.POST["farm"],
            source=source,
        )
        if created:
            logger.info("Importing %s from %s", farm.name, source)
            farm.refresh()
            messages.info(request, "Refreshed hosts")
        project.farm = farm
        project.save()
        return HttpResponseRedirect(reverse("project-detail", args=[project.id]))


class ExporterRegister(LoginRequiredMixin, FormView, mixins.ProjectMixin):
    model = models.Exporter
    template_name = "promgen/exporter_form.html"
    form_class = forms.ExporterForm

    def form_valid(self, form):
        project = get_object_or_404(models.Project, id=self.kwargs["pk"])
        exporter, _ = models.Exporter.objects.get_or_create(project=project, **form.clean())
        return HttpResponseRedirect(reverse("project-detail", args=[project.id]))


class ExporterScrape(LoginRequiredMixin, View):
    # TODO: Move to /rest/project/<slug>/scrape
    def post(self, request, pk):
        # Lookup our farm for testing
        farm = get_object_or_404(models.Project, pk=pk).farm

        # So we have a mutable dictionary
        data = request.POST.dict()

        # The default __metrics_path__ for Prometheus is /metrics so we need to
        # manually add it here in the case it's not set for our test
        if not data.setdefault("path", "/metrics"):
            data["path"] = "/metrics"

        def query():
            futures = []
            with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
                for host in farm.host_set.all():
                    futures.append(
                        executor.submit(
                            util.scrape,
                            "{scheme}://{host}:{port}{path}".format(host=host.name, **data),
                        )
                    )
                for future in concurrent.futures.as_completed(futures):
                    try:
                        result = future.result()
                        result.raise_for_status()
                        yield result.url, result.status_code
                    except requests.ConnectionError as e:
                        logger.warning("Error connecting to server")
                        yield e.request.url, "Error connecting to server"
                    except requests.RequestException as e:
                        logger.warning("Error with response")
                        yield e.request.url, str(e)
                    except Exception:
                        logger.exception("Unknown Exception")
                        yield "Unknown URL", "Unknown error"

        try:
            return JsonResponse(dict(query()))
        except Exception as e:
            return JsonResponse({"error": "Error with query %s" % e})


class URLRegister(LoginRequiredMixin, FormView, mixins.ProjectMixin):
    model = models.URL
    template_name = "promgen/url_form.html"
    form_class = forms.URLForm

    def form_valid(self, form):
        project = get_object_or_404(models.Project, id=self.kwargs["pk"])
        url, _ = models.URL.objects.get_or_create(project=project, **form.clean())
        return HttpResponseRedirect(reverse("project-detail", args=[project.id]))


class URLDelete(LoginRequiredMixin, DeleteView):
    model = models.URL

    def get_success_url(self):
        return reverse("project-detail", args=[self.object.project_id])


class URLList(LoginRequiredMixin, ListView):
    queryset = models.URL.objects.prefetch_related(
        "project",
        "project__service",
        "project__shard",
        "probe",
    )


class ProjectRegister(LoginRequiredMixin, CreateView):
    button_label = _("Register Project")
    model = models.Project
    fields = ["name", "description", "owner", "shard"]

    def get_initial(self):
        initial = {"owner": self.request.user}
        if "shard" in self.request.GET:
            initial["shard"] = get_object_or_404(models.Shard, pk=self.request.GET["shard"])
        return initial

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["service"] = get_object_or_404(models.Service, id=self.kwargs["pk"])
        context["shard_list"] = models.Shard.objects.all()
        return context

    def form_valid(self, form):
        form.instance.service_id = self.kwargs["pk"]
        return super().form_valid(form)


class ProjectUpdate(LoginRequiredMixin, UpdateView):
    model = models.Project
    button_label = _("Project Update")
    template_name = "promgen/project_form.html"
    fields = ["name", "description", "owner", "service", "shard"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["service"] = self.object.service
        context["shard_list"] = models.Shard.objects.all()
        return context


class ServiceUpdate(LoginRequiredMixin, UpdateView):
    button_label = _("Update Service")
    form_class = forms.ServiceUpdate
    model = models.Service


class RuleDetail(LoginRequiredMixin, DetailView):
    queryset = models.Rule.objects.prefetch_related(
        "content_object",
        "content_type",
        "overrides",
        "overrides__content_object",
        "overrides__content_type",
    )


class RuleUpdate(mixins.PromgenPermissionMixin, UpdateView):
    def get_permission_denied_message(self):
        return "Unable to edit rule %s. User lacks permission" % self.object

    def get_permission_required(self):
        # In the case of rules, we want to make sure the user has permission
        # to change the rule itself, but also permission to change the linked object
        self.object = self.get_object()
        obj = self.object._meta
        tgt = self.object.content_object._meta

        yield f"{obj.app_label}.change_{obj.model_name}"
        yield f"{tgt.app_label}.change_{tgt.model_name}"

    queryset = models.Rule.objects.prefetch_related(
        "content_object", "overrides", "overrides__content_object"
    )
    template_name = "promgen/rule_update.html"
    form_class = forms.AlertRuleForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["macro"] = macro.EXCLUSION_MACRO
        context["rules"] = [self.object.parent] if self.object.parent else [self.object]
        # We use setdefault here, because we may have received existing forms in the case of a POST
        # that has errors to be corrected
        context.setdefault("label_form", forms.LabelFormSet(initial=self.object.labels))
        context.setdefault(
            "annotation_form", forms.AnnotationFormSet(initial=self.object.annotations)
        )
        return context

    def form_invalid(self, **kwargs):
        """If the form is invalid, render the invalid form."""
        return self.render_to_response(self.get_context_data(**kwargs))

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()

        # Save a copy of our forms into a context var that we can use
        # to re-render our form properly in case of errors
        context = {}
        context["form"] = self.get_form()
        context["label_form"] = forms.LabelFormSet(data=request.POST)
        context["annotation_form"] = forms.AnnotationFormSet(data=request.POST)

        # Before we validate the form, we also want to move our labels and annotations
        # into our rule instance, so that they can be validated in the call to promtool
        context["form"].instance.labels = context["label_form"].to_dict()
        context["form"].instance.annotations = context["annotation_form"].to_dict()

        # With our labels+annotations manually cached we can test
        if not all(
            [
                context["form"].is_valid(),
                context["label_form"].is_valid(),
                context["annotation_form"].is_valid(),
            ]
        ):
            return self.form_invalid(**context)

        return self.form_valid(context["form"])


class AlertRuleRegister(mixins.PromgenPermissionMixin, mixins.RuleFormMixin, FormView):
    model = models.Rule
    template_name = "promgen/rule_register.html"
    form_class = forms.AlertRuleForm
    form_import_class = forms.ImportRuleForm

    def get_permission_required(self):
        # In the case of rules, we want to make sure the user has permission
        # to add the rule itself, but also permission to change the linked object
        yield "promgen.add_rule"
        yield "promgen.change_" + self.kwargs["content_type"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        # Set a dummy rule, so that our header/breadcrumbs render correctly
        context["rule"] = models.Rule()
        context["rule"].pk = 0
        context["rule"].set_object(self.kwargs["content_type"], self.kwargs["object_id"])
        context["macro"] = macro.EXCLUSION_MACRO
        return context

    def form_valid(self, form):
        form.instance.labels[form.instance.content_type.model] = form.instance.content_object.name
        form.instance.save()
        return HttpResponseRedirect(form.instance.get_absolute_url())

    def form_import(self, form, content_object):
        data = form.clean()
        counters = prometheus.import_rules_v2(data["rules"], content_object)
        messages.info(self.request, "Imported %s" % counters)
        return HttpResponseRedirect(content_object.get_absolute_url())


class ServiceRegister(LoginRequiredMixin, CreateView):
    button_label = _("Register Service")
    model = models.Service
    fields = ["name", "description", "owner"]

    def get_initial(self):
        return {"owner": self.request.user}


class FarmRegister(LoginRequiredMixin, FormView, mixins.ProjectMixin):
    model = models.Farm
    button_label = _("Register Farm")
    template_name = "promgen/farm_form.html"
    form_class = forms.FarmForm

    def form_valid(self, form):
        project = get_object_or_404(models.Project, id=self.kwargs["pk"])
        farm, _ = models.Farm.objects.get_or_create(source=discovery.FARM_DEFAULT, **form.clean())
        project.farm = farm
        project.save()
        return HttpResponseRedirect(project.get_absolute_url())


class ProjectNotifierRegister(LoginRequiredMixin, FormView, mixins.ProjectMixin):
    model = models.Sender
    template_name = "promgen/notifier_form.html"
    form_class = forms.SenderForm

    def form_valid(self, form):
        project = get_object_or_404(models.Project, id=self.kwargs["pk"])
        sender, created = models.Sender.objects.get_or_create(
            obj=project,
            **form.clean(),
            defaults={"owner": self.request.user},
        )
        signals.check_user_subscription(models.Sender, sender, created, self.request)
        return HttpResponseRedirect(project.get_absolute_url())


class ServiceNotifierRegister(LoginRequiredMixin, FormView, mixins.ServiceMixin):
    model = models.Sender
    template_name = "promgen/notifier_form.html"
    form_class = forms.SenderForm

    def form_valid(self, form):
        service = get_object_or_404(models.Service, id=self.kwargs["pk"])
        sender, created = models.Sender.objects.get_or_create(
            obj=service,
            **form.clean(),
            defaults={"owner": self.request.user},
        )
        signals.check_user_subscription(models.Sender, sender, created, self.request)
        return HttpResponseRedirect(service.get_absolute_url())


class SiteDetail(LoginRequiredMixin, TemplateView):
    template_name = "promgen/site_detail.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["rule_list"] = models.Rule.objects.filter(
            content_type__model="site", content_type__app_label="promgen"
        ).prefetch_related("content_object")
        return context


class Profile(LoginRequiredMixin, FormView):
    form_class = forms.SenderForm
    model = models.Sender
    template_name = "promgen/profile.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["discovery_plugins"] = [entry for entry in plugins.discovery()]
        context["notifier_plugins"] = [entry for entry in plugins.notifications()]
        context["notifiers"] = {"notifiers": models.Sender.objects.filter(obj=self.request.user)}
        context["subscriptions"] = models.Sender.objects.filter(
            sender="promgen.notification.user", value=self.request.user.username
        )
        return context

    def form_valid(self, form):
        sender, _ = models.Sender.objects.get_or_create(
            obj=self.request.user, owner=self.request.user, **form.clean()
        )
        return redirect("profile")


class HostRegister(LoginRequiredMixin, FormView):
    model = models.Host
    template_name = "promgen/host_form.html"
    form_class = forms.HostForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["farm"] = get_object_or_404(models.Farm, pk=self.kwargs["pk"])
        context["project"] = context["farm"].project_set.first()
        return context

    def form_valid(self, form):
        farm = get_object_or_404(models.Farm, id=self.kwargs["pk"])
        for hostname in form.cleaned_data["hosts"]:
            host, created = models.Host.objects.get_or_create(name=hostname, farm_id=farm.id)
            if created:
                logger.debug("Added %s to %s", host.name, farm.name)

        if farm.project_set.count() == 0:
            return redirect("farm-detail", pk=farm.id)
        return redirect("project-detail", pk=farm.project_set.first().id)


class ApiConfig(View):
    def get(self, request):
        return HttpResponse(prometheus.render_config(), content_type="application/json")

    def post(self, request, *args, **kwargs):
        try:
            body = json.loads(request.body.decode("utf-8"))

            prometheus.import_config(body, **kwargs)
        except Exception as e:
            return HttpResponse(e, status=400)

        return HttpResponse("Success", status=202)


class ApiQueue(View):
    def post(self, request):
        signals.trigger_write_config.send(request)
        signals.trigger_write_rules.send(request)
        signals.trigger_write_urls.send(request)
        return HttpResponse("OK", status=202)


class _ExportRules(View):
    def format(self, rules=None, name="promgen"):
        content = prometheus.render_rules(rules)
        response = HttpResponse(content)
        response["Content-Type"] = "application/x-yaml"
        response["Content-Disposition"] = "attachment; filename=%s.rule.yml" % name
        return response


class RulesConfig(_ExportRules):
    def get(self, request):
        return self.format()


class RuleExport(_ExportRules):
    def get(self, request, content_type, object_id):
        ct = ContentType.objects.get(
            app_label="promgen", model=content_type
        ).get_object_for_this_type(pk=object_id)
        rules = models.Rule.objects.filter(obj=ct)
        return self.format(rules)


class URLConfig(View):
    def get(self, request):
        return HttpResponse(prometheus.render_urls(), content_type="application/json")

    def post(self, request):
        tasks.write_urls()
        return HttpResponse("OK", status=202)


class AlertList(LoginRequiredMixin, ListView):
    paginate_by = 20
    queryset = models.Alert.objects.order_by("-created")

    def get_queryset(self):
        search = self.request.GET.get("search")
        if search:
            return self.queryset.filter(
                Q(alertlabel__name="Service", alertlabel__value__icontains=search)
                | Q(alertlabel__name="Project", alertlabel__value__icontains=search)
                | Q(alertlabel__name="Job", alertlabel__value__icontains=search)
            )

        qs = self.queryset
        for key, value in self.request.GET.items():
            if key in ["page", "search"]:
                continue
            elif key == "noSent":
                qs = qs.filter(sent_count=0)
            elif key == "sentError":
                qs = qs.exclude(error_count=0)
            else:
                qs = qs.filter(alertlabel__name=key, alertlabel__value=value)
        return qs


class AlertDetail(LoginRequiredMixin, DetailView):
    model = models.Alert

    def get_context_data(self, **kwargs):
        data = super().get_context_data(**kwargs)
        data["data"] = data["object"].json
        groupLabels = data["data"].get("groupLabels", {})
        commonLabels = data["data"].get("commonLabels", {})
        data["groupLabels"] = groupLabels
        data["otherLabels"] = {x: commonLabels[x] for x in commonLabels if x not in groupLabels}
        data["redirects"] = ["service", "project"]

        return data


class Metrics(View):
    def __init__(self):
        self.registry = prometheus_client.CollectorRegistry(auto_describe=True)
        prometheus_client.GCCollector(registry=self.registry)
        prometheus_client.PlatformCollector(registry=self.registry)
        prometheus_client.ProcessCollector(registry=self.registry)
        self.registry.register(self)

    def get(self, request, *args, **kwargs):
        return HttpResponse(
            prometheus_client.generate_latest(self.registry),
            content_type=prometheus_client.CONTENT_TYPE_LATEST,
        )

    def collect(self):
        # https://github.com/prometheus/client_python#custom-collectors
        v = GaugeMetricFamily(
            "promgen_build_info", "Promgen Information", labels=["version", "python"]
        )
        v.add_metric([settings.PROMGEN_VERSION, platform.python_version()], 1)
        yield v

        try:
            yield CounterMetricFamily(
                "promgen_alerts_processed",
                "Alerts",
                models.Alert.objects.latest("id").id,
            )
        except models.Alert.DoesNotExist:
            pass

        try:
            yield CounterMetricFamily(
                "promgen_alerts_failed",
                "Failed Alerts",
                models.AlertError.objects.latest("id").id,
            )
        except models.AlertError.DoesNotExist:
            pass

        yield GaugeMetricFamily("promgen_shards", "Registered Shards", models.Shard.objects.count())
        yield GaugeMetricFamily(
            "promgen_exporters", "Registered Exporters", models.Exporter.objects.count()
        )
        yield GaugeMetricFamily(
            "promgen_services", "Registered Services", models.Service.objects.count()
        )
        yield GaugeMetricFamily(
            "promgen_projects", "Registered Projects", models.Project.objects.count()
        )
        yield GaugeMetricFamily("promgen_rules", "Registered Rules", models.Rule.objects.count())
        yield GaugeMetricFamily("promgen_urls", "Registered URLs", models.URL.objects.count())

        # TODO Properly de-duplicate after refactoring
        yield GaugeMetricFamily(
            "promgen_hosts",
            "Registered Hosts",
            len(models.Host.objects.values("name").annotate(Count("name"))),
        )

        notifier = GaugeMetricFamily(
            "promgen_notifiers", "Registered Notifiers", labels=["type", "sender"]
        )
        for entry in models.Sender.objects.values("content_type__model", "sender").annotate(
            Count("sender"), count=Count("content_type")
        ):
            notifier.add_metric([entry["content_type__model"], entry["sender"]], entry["count"])

        yield notifier


class Search(LoginRequiredMixin, View):
    def get(self, request):
        MAPPING = {
            "farm_list": {
                "field": ("name__icontains",),
                "model": models.Farm,
                "prefetch": ("project_set", "host_set"),
                "query": ("search", "var-farm"),
            },
            "host_list": {
                "field": ("name__icontains",),
                "model": models.Host,
                "query": ("search", "var-instance"),
            },
            "project_list": {
                "field": ("name__icontains",),
                "model": models.Project,
                "prefetch": ("service", "notifiers", "exporter_set", "notifiers__owner"),
                "query": ("search", "var-project"),
            },
            "rule_list": {
                "field": ("name__icontains", "clause__icontains"),
                "model": models.Rule,
                "prefetch": ("content_object",),
                "query": ("search",),
            },
            "service_list": {
                "field": ("name__icontains",),
                "model": models.Service,
                "prefetch": ("project_set", "rule_set", "notifiers", "notifiers__owner"),
                "query": ("search", "var-service"),
            },
        }

        context = {}
        for target, obj in MAPPING.items():
            # If our potential search keys are not in our query string
            # then we can bail out quickly
            query = set(obj["query"]).intersection(request.GET.keys())
            if not query:
                logger.info("query for %s: <skipping>", target)
                continue
            logger.info("query for %s: %s", target, query)

            qs = obj["model"].objects
            if "prefetch" in obj:
                qs = qs.prefetch_related(*obj["prefetch"])

            # Build our OR query by combining Q lookups
            filters = None
            for var in query:
                for field in obj["field"]:
                    if filters:
                        filters |= Q(**{field: request.GET[var]})
                    else:
                        filters = Q(**{field: request.GET[var]})
            logger.info("filtering %s by %s", target, filters)

            qs = qs.filter(filters)
            context[target] = qs

        return render(request, "promgen/search.html", context)


class RuleImport(mixins.PromgenPermissionMixin, FormView):
    form_class = forms.ImportRuleForm
    template_name = "promgen/rule_import.html"

    # Since rule imports can change a lot of site wide stuff we
    # require site edit permission here
    permission_required = ("promgen.change_site", "promgen.change_rule")
    permisison_denied_message = "User lacks permission to import"

    def form_valid(self, form):
        data = form.clean()
        if data.get("file_field"):
            rules = data["file_field"].read().decode("utf8")
        elif data.get("rules"):
            rules = data.get("rules")
        else:
            messages.warning(self.request, "Missing rules")
            return self.form_invalid(form)

        try:
            counters = prometheus.import_rules_v2(rules)
            messages.info(self.request, "Imported %s" % counters)
            return redirect("rule-import")
        except Exception:
            messages.error(self.request, "Error importing rules")
            return self.form_invalid(form)


class Import(mixins.PromgenPermissionMixin, FormView):
    template_name = "promgen/import_form.html"
    form_class = forms.ImportConfigForm

    # Since imports can change a lot of site wide stuff we
    # require site edit permission here
    permission_required = ("promgen.change_site", "promgen.change_rule", "promgen.change_exporter")

    permission_denied_message = "User lacks permission to import"

    def form_valid(self, form):
        data = form.clean()
        if data.get("file_field"):
            messages.info(self.request, "Importing config from file")
            config = data["file_field"].read().decode("utf8")
        elif data.get("url"):
            messages.info(self.request, "Importing config from url")
            response = util.get(data["url"])
            response.raise_for_status()
            config = response.text
        elif data.get("config"):
            messages.info(self.request, "Importing config")
            config = data["config"]
        else:
            messages.warning(self.request, "Missing config")
            return self.form_invalid(form)

        kwargs = {}
        # This also lets us catch passing an empty string to signal using
        # the shard value from the post request
        if data.get("shard"):
            kwargs["replace_shard"] = data.get("shard")

        imported, skipped = prometheus.import_config(json.loads(config), **kwargs)

        if imported:
            counters = {key: len(imported[key]) for key in imported}
            messages.info(self.request, "Imported %s" % counters)

        if skipped:
            counters = {key: len(skipped[key]) for key in skipped}
            messages.info(self.request, "Skipped %s" % counters)

        # If we only have a single object in a category, automatically
        # redirect to that category to make things easier to understand
        if len(imported["Project"]) == 1:
            return HttpResponseRedirect(imported["Project"][0].get_absolute_url())
        if len(imported["Service"]) == 1:
            return HttpResponseRedirect(imported["Service"][0].get_absolute_url())
        if len(imported["Shard"]) == 1:
            return HttpResponseRedirect(imported["Shard"][0].get_absolute_url())

        return redirect("service-list")


class RuleTest(LoginRequiredMixin, View):
    def post(self, request, pk):
        if pk == 0:
            rule = models.Rule()
            rule.set_object(request.POST["content_type"], request.POST["object_id"])
        else:
            rule = get_object_or_404(models.Rule, id=pk)

        # Default values in case we do not have a more specific content_type to test against
        expected_labels = {
            "service": set(),
            "project": set(),
        }
        unexpected_labels = collections.defaultdict(set)

        # Given our current rule, we want to see what service/project it will affect, so that we can
        # check our test query output for labels outside our expected match
        if rule.content_type.model == "service":
            # for a service rule, we expect the current service and all child projects as expected
            expected_labels["service"] = {rule.content_object.name}
            expected_labels["project"] = {
                project.name for project in rule.content_object.project_set.all()
            }

        if rule.content_type.model == "project":
            # for a project rule we expect only the current project and the parent service
            expected_labels["service"] = {rule.content_object.service.name}
            expected_labels["project"] = {rule.content_object.name}

        query = macro.rulemacro(rule, request.POST["query"])
        # Since our rules affect all servers we use Promgen's proxy-query to test our rule
        # against all the servers at once
        url = resolve_domain("proxy-query")

        logger.debug("Querying %s with %s", url, query)
        start = time.time()
        result = util.get(url, {"query": query}).json()
        result["duration"] = datetime.timedelta(seconds=(time.time() - start))
        result["query"] = query

        # TODO: This could be more robust, but for now this ensures failed queries
        # without a 'data' field get processed
        metrics = result.get("data", {}).setdefault("result", [])
        result["collapse"] = len(metrics) > 5
        errors = result.setdefault("errors", {})

        if len(metrics) == 0:
            errors["no_results"] = (
                "No results. "
                "You may need to temporarily remove conditional checks (> < ==) to verify."
            )

        for row in metrics:
            if "service" not in row["metric"] and "project" not in row["metric"]:
                errors["routing"] = (
                    "Some metrics are missing service and project labels. "
                    "Promgen will be unable to route message."
                )

            for label in expected_labels:
                if label in row["metric"]:
                    if row["metric"][label] not in expected_labels[label]:
                        unexpected_labels[label].add(row["metric"][label])

        # For each key in our unexpected_labels, we want to go through the returned
        # values and check to see if it is an expected one or not.
        for label in unexpected_labels:
            if unexpected_labels[label]:
                result["severity"] = "danger"
                errors[f"unrelated_{label}"] = (
                    f"Unrelated {label} labels found\n"
                    f"Expected: {expected_labels[label]}\n"
                    f"Found: {unexpected_labels[label]}"
                )

        # Place this at the bottom to have a query error show up as danger
        if result["status"] != "success":
            result["severity"] = "danger"
            errors["Query"] = result["error"]

        return JsonResponse(
            {request.POST["target"]: render_to_string("promgen/ajax_clause_check.html", result)}
        )



# Copyright (c) 2018 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

from django.core.management.base import BaseCommand

from promgen.models import DefaultExporter


class Command(BaseCommand):
    help = """Register default exporter from the commandline"""

    # This is intended to be used from a configuration management tool
    # where there may already be a port mapping that we want to import
    # into Promgen

    def add_arguments(self, parser):
        parser.add_argument("job")
        parser.add_argument("port", type=int)
        parser.add_argument("path", nargs="?", default="")

    def handle(self, job, port, path, **kargs):
        exporter, created = DefaultExporter.objects.get_or_create(job=job, port=port, path=path)
        if created:
            self.stdout.write(f"Created {exporter}")
        else:
            self.stdout.write(f"Already exists {exporter}")



import django_filters


class ShardFilter(django_filters.rest_framework.FilterSet):
    name = django_filters.CharFilter(field_name="name", lookup_expr="contains")


class ServiceFilter(django_filters.rest_framework.FilterSet):
    name = django_filters.CharFilter(field_name="name", lookup_expr="contains")


class ProjectFilter(django_filters.rest_framework.FilterSet):
    name = django_filters.CharFilter(field_name="name", lookup_expr="contains")
    service = django_filters.CharFilter(field_name="service__name", lookup_expr="contains")
    shard = django_filters.CharFilter(field_name="shard__name", lookup_expr="contains")


class RuleFilter(django_filters.rest_framework.FilterSet):
    name = django_filters.CharFilter(field_name="name", lookup_expr="contains")
    parent = django_filters.CharFilter(field_name="parent__name", lookup_expr="contains")
    enabled = django_filters.BooleanFilter(field_name="enabled")



# Copyright (c) 2018 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

"""
Prune old alerts from Promgen's Database

Simple command to prune old alerts from Promgen's Database
based on days.

Use without arguments as dryrun or --force to execute
"""

import datetime

from django.core.management.base import BaseCommand
from django.utils import timezone

from promgen import models


class Command(BaseCommand):
    help = __doc__.strip().split("\n")[0]

    def add_arguments(self, parser):
        parser.add_argument("--days", type=int, default=30, help="Days of alerts to delete")
        parser.add_argument(
            "--force",
            dest="dryrun",
            action="store_false",
            help="Defaults to dry run. Use to execute operation",
        )

    def success(self, message, *args):
        self.stdout.write(self.style.SUCCESS(message % args))

    def handle(self, days, dryrun, verbosity, **options):
        cutoff = timezone.now() - datetime.timedelta(days=days)

        if verbosity > 1:
            self.success("Removing alerts before %s (%d days)", cutoff, days)

        alerts = models.Alert.objects.filter(created__lt=cutoff)

        if dryrun:
            self.success("Would have removed %d alerts", alerts.count())
            return

        count, objects = alerts.delete()

        if verbosity > 1:
            self.success("Removed %d Alerts", count)



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import os
import pathlib

# Django environment defaults
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "promgen.settings")
default_app_config = "promgen.apps.PromgenConfig"

# Promgen environment defaults
PROMGEN_CONFIG_DIR = pathlib.Path(
    os.environ.setdefault("PROMGEN_CONFIG_DIR", os.path.expanduser("~/.config/promgen"))
)

PROMGEN_CONFIG_FILE = pathlib.Path(
    os.environ.get("PROMGEN_CONFIG", str(PROMGEN_CONFIG_DIR / "promgen.yml"))
)



# Copyright (c) 2020 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import time

from django.core.management.base import BaseCommand

from promgen import models, tasks


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--force",
            dest="dryrun",
            action="store_false",
            help="Defaults to dry run. Use to execute operation",
        )

    def handle(self, dryrun, **kargs):
        for alert in models.Alert.objects.filter(alertlabel__isnull=True):
            if dryrun:
                labels = alert.json.get("commonLabels")
                self.stderr.write(f"alert_id: {alert.pk}, labels: {labels}")
                continue

            tasks.index_alert.delay(alert.pk)
            time.sleep(0.1)



# Copyright (c) 2020 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE
# Generated by Django 2.2.9 on 2020-01-16 06:35

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("promgen", "0011_notifier_counts")]

    operations = [
        migrations.AlterField(
            model_name="farm",
            name="name",
            field=models.CharField(
                max_length=128,
                validators=[
                    django.core.validators.RegexValidator(
                        "^[0-9a-zA-Z_]*$", "Only alphanumeric characters are allowed."
                    )
                ],
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="name",
            field=models.CharField(
                max_length=128,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        "^[0-9a-zA-Z_]*$", "Only alphanumeric characters are allowed."
                    )
                ],
            ),
        ),
        migrations.AlterField(
            model_name="service",
            name="name",
            field=models.CharField(
                max_length=128,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        "^[0-9a-zA-Z_]*$", "Only alphanumeric characters are allowed."
                    )
                ],
            ),
        ),
        migrations.AlterField(
            model_name="shard",
            name="name",
            field=models.CharField(
                max_length=128,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        "^[0-9a-zA-Z_]*$", "Only alphanumeric characters are allowed."
                    )
                ],
            ),
        ),
    ]



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE


import logging
import socket

import celery

from celery.signals import celeryd_after_setup


logger = logging.getLogger(__name__)


app = celery.Celery("promgen")

# Using a string here means the worker don't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object("django.conf:settings", namespace="CELERY")

# Load task modules from all registered Django app configs.
app.autodiscover_tasks()


@app.task(bind=True)
def debug_task(self):
    print(f"Request: {self.request!r}")


@celeryd_after_setup.connect
def setup_direct_queue(sender, instance, **kwargs):
    # To enable triggering config writes and reloads on a specific Prometheus server
    # we automatically create a queue for the current server that we can target from
    # our promgen.prometheus functions
    instance.app.amqp.queues.select_add(socket.gethostname())
    debug_task.apply_async(queue=socket.gethostname())



from enum import Enum
from django.utils.translation import gettext as _
from django.core.exceptions import ValidationError


class ValidationErrorEnum(Enum):
    def error(self, **params):
        params.setdefault("message", self.value)
        params.setdefault("code", self.name)
        return ValidationError(**params)


class SilenceError(ValidationErrorEnum):
    NOLABEL = _("Missing labels for silence.")
    GLOBALSILENCE = _("Unable to silence global rules with alertname alone.")
    STARTENDTIME = _("Both start and end are required.")
    STARTENDMISMATCH = _("Start time and end time are mismatched.")



# Copyright (c) 2017 LINE Corporation
# These sources are released under the terms of the MIT license: see LICENSE

import re
from functools import partial

from dateutil import parser
from django import forms
from django.core.exceptions import ValidationError

from promgen import models, plugins, prometheus, validators, errors


class ImportConfigForm(forms.Form):
    def _choices():
        return [("", "<Default>")] + sorted(
            (shard.name, "Import into: " + shard.name) for shard in models.Shard.objects.all()
        )

    config = forms.CharField(
        widget=forms.Textarea(attrs={"rows": 5, "class": "form-control"}), required=False
    )
    url = forms.CharField(widget=forms.TextInput(attrs={"class": "form-control"}), required=False)
    file_field = forms.FileField(
        widget=forms.FileInput(attrs={"class": "form-control"}), required=False
    )

    shard = forms.ChoiceField(choices=_choices, required=False)


class ImportRuleForm(forms.Form):
    rules = forms.CharField(
        widget=forms.Textarea(attrs={"rows": 5, "class": "form-control"}), required=False
    )
    file_field = forms.FileField(
        widget=forms.FileInput(attrs={"class": "form-control"}), required=False
    )

    def clean(self):
        if any(self.cleaned_data.values()):
            return self.cleaned_data
        raise forms.ValidationError("Must submit data or file")


class SilenceForm(forms.Form):
    duration = forms.CharField(required=False, validators=[validators.duration])
    startsAt = forms.CharField(required=False, validators=[validators.datetime])
    endsAt = forms.CharField(required=False, validators=[validators.datetime])
    comment = forms.CharField(required=False)
    createdBy = forms.CharField(required=False)

    def clean_comment(self):
        if self.cleaned_data["comment"]:
            return self.cleaned_data["comment"]
        return "Silenced from Promgen"

    def clean_createdBy(self):
        if self.cleaned_data["createdBy"]:
            return self.cleaned_data["createdBy"]
        return "Promgen"

    def clean(self):
        data = super().clean()

        # Validation for labels
        if "labels" not in self.data or not self.data["labels"]:
            raise errors.SilenceError.NOLABEL.error()

        # Users should not be able to accidentally silence a global rule without
        # setting some other labels as well.
        if "alertname" in self.data["labels"]:
            if "service" not in self.data["labels"] and "project" not in self.data["labels"]:
                rule = models.Rule.objects.get(name=self.data["labels"]["alertname"])
                if rule.content_type.model == "site":
                    raise errors.SilenceError.GLOBALSILENCE.error()

        # Once labels have been validated, we want to add them to our cleaned data so
        # they can be submitted.
        self.cleaned_data["labels"] = self.data["labels"]

        if data.get("duration"):
            # No further validation is required if only duration is set
            return

        # Validate our start/end times
        start = data.get("startsAt")
        stop = data.get("endsAt")

        if not all([start, stop]):
            raise errors.SilenceError.STARTENDTIME.error()

        elif parser.parse(start) > parser.parse(stop):
            raise errors.SilenceError.STARTENDMISMATCH.error()


class SilenceExpireForm(forms.Form):
    silence_id = forms.CharField(required=True)
    next = forms.CharField(required=False)


class ExporterForm(forms.ModelForm):
    class Meta:
        model = models.Exporter
        exclude = ["project"]

        widgets = {
            "job": forms.TextInput(attrs={"class": "form-control"}),
            "port": forms.TextInput(attrs={"class": "form-control", "type": "number"}),
            "path": forms.TextInput(attrs={"class": "form-control", "placeholder": "/metrics"}),
            "scheme": forms.Select(attrs={"class": "form-control"}),
        }


class ServiceRegister(forms.ModelForm):
    class Meta:
        model = models.Service
        # shard is determined by the pk in the service register url
        exclude = ["shard"]


class ServiceUpdate(forms.ModelForm):
    class Meta:
        model = models.Service
        exclude = []


class URLForm(forms.ModelForm):
    class Meta:
        model = models.URL
        fields = ["url", "probe"]
        widgets = {
            "url": forms.TextInput(attrs={"class": "form-control input-sm"}),
            "probe": forms.Select(attrs={"class": "form-control input-sm"}),
        }


class AlertRuleForm(forms.ModelForm):
    class Meta:
        model = models.Rule
        widgets = {
            "name": forms.TextInput(attrs={"class": "form-control"}),
            "duration": forms.TextInput(attrs={"class": "form-control"}),
            "clause": forms.Textarea(attrs={"rows": 5, "class": "form-control"}),
            "enabled": forms.CheckboxInput(attrs={"data-toggle": "toggle", "data-size": "mini"}),
            "description": forms.Textarea(attrs={"rows": 5, "class": "form-control"}),
        }
        # We define a custom widget for each of our fields, so we just take the
        # keys here to avoid manually updating a list of fields.
        fields = widgets.keys()

    def clean(self):
        # Check our cleaned data then let Prometheus check our rule
        super().clean()
        rule = models.Rule(**self.cleaned_data)

        # Make sure we pull in our labels and annotations for
        # testing if needed
        # See django docs on cached_property
        rule.labels = self.instance.labels
        rule.annotations = self.instance.annotations

        prometheus.check_rules([rule])


class _KeyValueForm(forms.Form):
    key = forms.CharField(widget=forms.TextInput(attrs={"class": "form-control"}))
    value = forms.CharField(widget=forms.TextInput(attrs={"class": "form-control"}))


# We need a custom KeyValueSet because we need to be able to convert between the single dictionary
# form saved to our models, and the list of models used by
class _KeyValueSet(forms.BaseFormSet):
    def __init__(self, initial=None, **kwargs):
        if initial:
            kwargs["initial"] = [{"key": key, "value": initial[key]} for key in initial]
        super().__init__(**kwargs, form_kwargs={"empty_permitted": True})

    def to_dict(self):
        return {x["key"]: x["value"] for x in self.cleaned_data if x and not x["DELETE"]}


# For both LabelFormSet and AnnotationFormSet we always want to have a prefix assigned, but it's
# awkward if we need to specify it in multiple places. We use a partial here, so that it is the same
# as always passing prefix as part of our __init__ call.
LabelFormSet = partial(
    forms.formset_factory(
        form=_KeyValueForm,
        formset=_KeyValueSet,
        can_delete=True,
        extra=1,
    ),
    prefix="labels",
)

AnnotationFormSet = partial(
    forms.formset_factory(
        form=_KeyValueForm,
        formset=_KeyValueSet,
        can_delete=True,
        extra=1,
    ),
    prefix="annotations",
)


class RuleCopyForm(forms.Form):
    content_type = forms.ChoiceField(choices=[(x, x) for x in ["service", "project"]])
    object_id = forms.IntegerField()


class FarmForm(forms.ModelForm):
    class Meta:
        model = models.Farm
        exclude = ["source"]


class SenderForm(forms.ModelForm):
    sender = forms.ChoiceField(
        choices=[(entry.module_name, entry.module_name) for entry in plugins.notifications()]
    )

    class Meta:
        model = models.Sender
        exclude = ["content_type", "object_id", "owner", "enabled"]


class NotifierUpdate(forms.ModelForm):
    class Meta:
        model = models.Sender
        exclude = ["value"]


class HostForm(forms.Form):
    hosts = forms.CharField(widget=forms.Textarea)

    def clean(self):
        # Hosts may be submitted delimiated by either , or newline so
        # once we split on that, we want to make sure there are no invalid
        # hostnames
        hosts = set()
        for hostname in re.split(r"[,\s]+", self.cleaned_data["hosts"]):
            if hostname == "":
                continue
            validators.hostname(hostname)
            hosts.add(hostname)
        if not hosts:
            raise ValidationError("No valid hosts")
        self.cleaned_data["hosts"] = list(hosts)
